# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/stats/kickoff.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ...api import player_id_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/stats/kickoff.proto',
  package='api.stats',
  serialized_pb=_b('\n\x17\x61pi/stats/kickoff.proto\x12\tapi.stats\x1a\x13\x61pi/player_id.proto\"A\n\x12PlayerPositionData\x12\r\n\x05pos_x\x18\x01 \x01(\x02\x12\r\n\x05pos_y\x18\x02 \x01(\x02\x12\r\n\x05pos_z\x18\x03 \x01(\x02\"\xde\x02\n\rKickoffPlayer\x12\x1d\n\x06player\x18\x01 \x01(\x0b\x32\r.api.PlayerId\x12\x34\n\x10kickoff_position\x18\x02 \x01(\x0e\x32\x1a.api.stats.KickoffPosition\x12\x30\n\x0etouch_position\x18\x03 \x01(\x0e\x32\x18.api.stats.TouchPosition\x12\x36\n\x0fplayer_position\x18\x04 \x01(\x0b\x32\x1d.api.stats.PlayerPositionData\x12\r\n\x05\x62oost\x18\x05 \x01(\x02\x12\r\n\x05jumps\x18\x06 \x03(\x02\x12\x11\n\tball_dist\x18\x07 \x01(\x02\x12\x12\n\nboost_time\x18\x08 \x01(\x02\x12\x12\n\nstart_left\x18\t \x01(\x08\x12\x35\n\x0estart_position\x18\n \x01(\x0b\x32\x1d.api.stats.PlayerPositionData\"\x8d\x01\n\x05Touch\x12)\n\x07players\x18\x01 \x03(\x0b\x32\x18.api.stats.KickoffPlayer\x12\x14\n\x0ckickoff_goal\x18\x02 \x01(\x02\x12\x18\n\x10orange_team_goal\x18\x03 \x01(\x08\x12)\n\x12\x66irst_touch_player\x18\x04 \x01(\x0b\x32\r.api.PlayerId\"\x93\x01\n\x0cKickoffStats\x12\x13\n\x0bstart_frame\x18\x01 \x01(\x05\x12\x13\n\x0btouch_frame\x18\x02 \x01(\x05\x12\x12\n\ntouch_time\x18\x03 \x01(\x02\x12$\n\x04type\x18\x04 \x01(\x0e\x32\x16.api.stats.KickoffType\x12\x1f\n\x05touch\x18\x05 \x01(\x0b\x32\x10.api.stats.Touch*S\n\x0fKickoffPosition\x12\x0c\n\x08\x44IAGONAL\x10\x00\x12\r\n\tOFFCENTER\x10\x01\x12\n\n\x06GOALIE\x10\x02\x12\x17\n\x13UNKNOWN_KICKOFF_POS\x10\x03*\xe0\x02\n\x0bKickoffType\x12\x1c\n\x18THREES_DIAG_DIAG_OFFCENT\x10\x00\x12\x19\n\x15THREES_DIAG_DIAG_GOAL\x10\x01\x12\x1f\n\x1bTHREES_DIAG_OFFCENT_OFFCENT\x10\x02\x12\x1c\n\x18THREES_DIAG_OFFCENT_GOAL\x10\x03\x12\x1f\n\x1bTHREES_OFFCENT_OFFCENT_GOAL\x10\x04\x12\x12\n\x0eTWOS_DIAG_DIAG\x10\x05\x12\x15\n\x11TWOS_DIAG_OFFCENT\x10\x06\x12\x18\n\x14TWOS_OFFCENT_OFFCENT\x10\x07\x12\x15\n\x11TWOS_OFFCENT_GOAL\x10\x08\x12\x12\n\x0eTWOS_DIAG_GOAL\x10\t\x12\r\n\tDUEL_DIAG\x10\n\x12\x10\n\x0c\x44UEL_OFFCENT\x10\x0b\x12\r\n\tDUEL_GOAL\x10\x0c\x12\x18\n\x14UNKNOWN_KICKOFF_TYPE\x10\r*Y\n\rTouchPosition\x12\t\n\x05\x42OOST\x10\x00\x12\x08\n\x04\x42\x41LL\x10\x01\x12\x07\n\x03\x41\x46K\x10\x02\x12\x08\n\x04GOAL\x10\x03\x12\t\n\x05\x43HEAT\x10\x04\x12\x15\n\x11UNKNOWN_TOUCH_POS\x10\x05')
  ,
  dependencies=[player_id_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_KICKOFFPOSITION = _descriptor.EnumDescriptor(
  name='KickoffPosition',
  full_name='api.stats.KickoffPosition',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DIAGONAL', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OFFCENTER', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GOALIE', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_KICKOFF_POS', index=3, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=773,
  serialized_end=856,
)
_sym_db.RegisterEnumDescriptor(_KICKOFFPOSITION)

KickoffPosition = enum_type_wrapper.EnumTypeWrapper(_KICKOFFPOSITION)
_KICKOFFTYPE = _descriptor.EnumDescriptor(
  name='KickoffType',
  full_name='api.stats.KickoffType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='THREES_DIAG_DIAG_OFFCENT', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='THREES_DIAG_DIAG_GOAL', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='THREES_DIAG_OFFCENT_OFFCENT', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='THREES_DIAG_OFFCENT_GOAL', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='THREES_OFFCENT_OFFCENT_GOAL', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TWOS_DIAG_DIAG', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TWOS_DIAG_OFFCENT', index=6, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TWOS_OFFCENT_OFFCENT', index=7, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TWOS_OFFCENT_GOAL', index=8, number=8,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TWOS_DIAG_GOAL', index=9, number=9,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DUEL_DIAG', index=10, number=10,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DUEL_OFFCENT', index=11, number=11,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DUEL_GOAL', index=12, number=12,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_KICKOFF_TYPE', index=13, number=13,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=859,
  serialized_end=1211,
)
_sym_db.RegisterEnumDescriptor(_KICKOFFTYPE)

KickoffType = enum_type_wrapper.EnumTypeWrapper(_KICKOFFTYPE)
_TOUCHPOSITION = _descriptor.EnumDescriptor(
  name='TouchPosition',
  full_name='api.stats.TouchPosition',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='BOOST', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BALL', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AFK', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GOAL', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CHEAT', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_TOUCH_POS', index=5, number=5,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1213,
  serialized_end=1302,
)
_sym_db.RegisterEnumDescriptor(_TOUCHPOSITION)

TouchPosition = enum_type_wrapper.EnumTypeWrapper(_TOUCHPOSITION)
DIAGONAL = 0
OFFCENTER = 1
GOALIE = 2
UNKNOWN_KICKOFF_POS = 3
THREES_DIAG_DIAG_OFFCENT = 0
THREES_DIAG_DIAG_GOAL = 1
THREES_DIAG_OFFCENT_OFFCENT = 2
THREES_DIAG_OFFCENT_GOAL = 3
THREES_OFFCENT_OFFCENT_GOAL = 4
TWOS_DIAG_DIAG = 5
TWOS_DIAG_OFFCENT = 6
TWOS_OFFCENT_OFFCENT = 7
TWOS_OFFCENT_GOAL = 8
TWOS_DIAG_GOAL = 9
DUEL_DIAG = 10
DUEL_OFFCENT = 11
DUEL_GOAL = 12
UNKNOWN_KICKOFF_TYPE = 13
BOOST = 0
BALL = 1
AFK = 2
GOAL = 3
CHEAT = 4
UNKNOWN_TOUCH_POS = 5



_PLAYERPOSITIONDATA = _descriptor.Descriptor(
  name='PlayerPositionData',
  full_name='api.stats.PlayerPositionData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pos_x', full_name='api.stats.PlayerPositionData.pos_x', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pos_y', full_name='api.stats.PlayerPositionData.pos_y', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pos_z', full_name='api.stats.PlayerPositionData.pos_z', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=59,
  serialized_end=124,
)


_KICKOFFPLAYER = _descriptor.Descriptor(
  name='KickoffPlayer',
  full_name='api.stats.KickoffPlayer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='player', full_name='api.stats.KickoffPlayer.player', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='kickoff_position', full_name='api.stats.KickoffPlayer.kickoff_position', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='touch_position', full_name='api.stats.KickoffPlayer.touch_position', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='player_position', full_name='api.stats.KickoffPlayer.player_position', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='boost', full_name='api.stats.KickoffPlayer.boost', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='jumps', full_name='api.stats.KickoffPlayer.jumps', index=5,
      number=6, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ball_dist', full_name='api.stats.KickoffPlayer.ball_dist', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='boost_time', full_name='api.stats.KickoffPlayer.boost_time', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='start_left', full_name='api.stats.KickoffPlayer.start_left', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='start_position', full_name='api.stats.KickoffPlayer.start_position', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=127,
  serialized_end=477,
)


_TOUCH = _descriptor.Descriptor(
  name='Touch',
  full_name='api.stats.Touch',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='players', full_name='api.stats.Touch.players', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='kickoff_goal', full_name='api.stats.Touch.kickoff_goal', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='orange_team_goal', full_name='api.stats.Touch.orange_team_goal', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='first_touch_player', full_name='api.stats.Touch.first_touch_player', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=480,
  serialized_end=621,
)


_KICKOFFSTATS = _descriptor.Descriptor(
  name='KickoffStats',
  full_name='api.stats.KickoffStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='start_frame', full_name='api.stats.KickoffStats.start_frame', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='touch_frame', full_name='api.stats.KickoffStats.touch_frame', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='touch_time', full_name='api.stats.KickoffStats.touch_time', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='api.stats.KickoffStats.type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='touch', full_name='api.stats.KickoffStats.touch', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=624,
  serialized_end=771,
)

_KICKOFFPLAYER.fields_by_name['player'].message_type = player_id_pb2._PLAYERID
_KICKOFFPLAYER.fields_by_name['kickoff_position'].enum_type = _KICKOFFPOSITION
_KICKOFFPLAYER.fields_by_name['touch_position'].enum_type = _TOUCHPOSITION
_KICKOFFPLAYER.fields_by_name['player_position'].message_type = _PLAYERPOSITIONDATA
_KICKOFFPLAYER.fields_by_name['start_position'].message_type = _PLAYERPOSITIONDATA
_TOUCH.fields_by_name['players'].message_type = _KICKOFFPLAYER
_TOUCH.fields_by_name['first_touch_player'].message_type = player_id_pb2._PLAYERID
_KICKOFFSTATS.fields_by_name['type'].enum_type = _KICKOFFTYPE
_KICKOFFSTATS.fields_by_name['touch'].message_type = _TOUCH
DESCRIPTOR.message_types_by_name['PlayerPositionData'] = _PLAYERPOSITIONDATA
DESCRIPTOR.message_types_by_name['KickoffPlayer'] = _KICKOFFPLAYER
DESCRIPTOR.message_types_by_name['Touch'] = _TOUCH
DESCRIPTOR.message_types_by_name['KickoffStats'] = _KICKOFFSTATS
DESCRIPTOR.enum_types_by_name['KickoffPosition'] = _KICKOFFPOSITION
DESCRIPTOR.enum_types_by_name['KickoffType'] = _KICKOFFTYPE
DESCRIPTOR.enum_types_by_name['TouchPosition'] = _TOUCHPOSITION

PlayerPositionData = _reflection.GeneratedProtocolMessageType('PlayerPositionData', (_message.Message,), dict(
  DESCRIPTOR = _PLAYERPOSITIONDATA,
  __module__ = 'api.stats.kickoff_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.PlayerPositionData)
  ))
_sym_db.RegisterMessage(PlayerPositionData)

KickoffPlayer = _reflection.GeneratedProtocolMessageType('KickoffPlayer', (_message.Message,), dict(
  DESCRIPTOR = _KICKOFFPLAYER,
  __module__ = 'api.stats.kickoff_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.KickoffPlayer)
  ))
_sym_db.RegisterMessage(KickoffPlayer)

Touch = _reflection.GeneratedProtocolMessageType('Touch', (_message.Message,), dict(
  DESCRIPTOR = _TOUCH,
  __module__ = 'api.stats.kickoff_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.Touch)
  ))
_sym_db.RegisterMessage(Touch)

KickoffStats = _reflection.GeneratedProtocolMessageType('KickoffStats', (_message.Message,), dict(
  DESCRIPTOR = _KICKOFFSTATS,
  __module__ = 'api.stats.kickoff_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.KickoffStats)
  ))
_sym_db.RegisterMessage(KickoffStats)


# @@protoc_insertion_point(module_scope)



class ApiPlayerLoadout:

    @staticmethod
    def create_from_player(proto_loadout, player):
        try:
            loadout = player.loadout[player.is_orange]
            paints = player.paint[player.is_orange]
            user_colors = player.user_colors[player.is_orange]
        except IndexError:
            return

        if loadout['banner'] is not None:
            proto_loadout.banner = loadout['banner']
        if loadout['boost'] is not None:
            proto_loadout.boost = loadout['boost']
        if loadout['car'] is not None:
            proto_loadout.car = loadout['car']
        if loadout['goal_explosion'] is not None:
            proto_loadout.goal_explosion = loadout['goal_explosion']
        if loadout['skin'] is not None:
            proto_loadout.skin = loadout['skin']
        if loadout['trail'] is not None:
            proto_loadout.trail = loadout['trail']
        if loadout['version'] is not None:
            proto_loadout.version = loadout['version']
        if loadout['wheels'] is not None:
            proto_loadout.wheels = loadout['wheels']
        if loadout['topper'] is not None:
            proto_loadout.topper = loadout['topper']
        if loadout['antenna'] is not None:
            proto_loadout.antenna = loadout['antenna']
        if loadout['engine_audio'] is not None:
            proto_loadout.engine_audio = loadout['engine_audio']
        if loadout['avatar_border'] is not None:
            proto_loadout.engine_audio = loadout['avatar_border']

        if paints['banner'] is not None:
            proto_loadout.banner_paint = paints['banner']
        if paints['boost'] is not None:
            proto_loadout.boost_paint = paints['boost']
        if paints['car'] is not None:
            proto_loadout.car_paint = paints['car']
        if paints['goal_explosion'] is not None:
            proto_loadout.goal_explosion_paint = paints['goal_explosion']
        if paints['skin'] is not None:
            proto_loadout.skin_paint = paints['skin']
        if paints['trail'] is not None:
            proto_loadout.trail_paint = paints['trail']
        if paints['wheels'] is not None:
            proto_loadout.wheels_paint = paints['wheels']
        if paints['topper'] is not None:
            proto_loadout.topper_paint = paints['topper']
        if paints['antenna'] is not None:
            proto_loadout.antenna_paint = paints['antenna']

        if loadout['primary_color'] is not None:
            proto_loadout.primary_color = loadout['primary_color']
        if loadout['accent_color'] is not None:
            proto_loadout.accent_color = loadout['accent_color']
        if loadout['primary_finish'] is not None:
            proto_loadout.primary_finish = loadout['primary_finish']
        if loadout['accent_finish'] is not None:
            proto_loadout.accent_finish = loadout['accent_finish']

        if user_colors['banner'] is not None:
            proto_loadout.banner_user_color = user_colors['banner']
        if user_colors['avatar_border'] is not None:
            proto_loadout.avatar_border_user_color = user_colors['avatar_border']



from typing import Dict

import pandas

from carball.analysis.stats.stats import HitStat
from carball.generated.api import game_pb2
from carball.generated.api.player_pb2 import Player
from carball.generated.api.stats import stats_pb2
from carball.generated.api.stats.events_pb2 import Hit
from carball.json_parser.game import Game


class HitCountStat(HitStat):

    def initialize_hit_stat(self, game: Game, player_map: Dict[str, Player], data_frame: pandas.DataFrame):
        pass

    def apply_stat(self, hit_count: stats_pb2.HitCounts, saltie_hit):
        if saltie_hit.dribble:
            hit_count.total_dribbles += 1
        if saltie_hit.dribble_continuation:
            hit_count.total_dribble_conts += 1
        if saltie_hit.pass_:
            hit_count.total_passes += 1
        if saltie_hit.goal:
            hit_count.total_goals += 1
        if saltie_hit.shot:
            hit_count.total_shots += 1
        if saltie_hit.save:
            hit_count.total_saves += 1
        if saltie_hit.aerial:
            hit_count.total_aerials += 1
        if saltie_hit.clear:
            hit_count.total_clears += 1
        hit_count.total_hits += 1

    def calculate_next_hit_stat(self, game: Game, proto_game: game_pb2.Game, saltie_hit: Hit, next_saltie_hit: Hit,
                                player_map: Dict[str, Player], hit_index: int):
        player = player_map[saltie_hit.player_id.id]
        self.apply_stat(player.stats.hit_counts, saltie_hit)
        team = proto_game.teams[player.is_orange]
        self.apply_stat(team.stats.hit_counts, saltie_hit)



from typing import Dict

import pandas as pd
from carball.generated.api.stats.kickoff_pb2 import KickoffPlayer, TouchPosition, BOOST, CHEAT, GOAL, BALL, AFK

from carball.analysis.stats.stats import BaseStat
from carball.generated.api import game_pb2
from carball.generated.api.player_pb2 import Player
from carball.generated.api.stats.player_stats_pb2 import PlayerStats, CumulativeKickoffStats
from carball.json_parser.game import Game


class KickoffStat(BaseStat):
    def calculate_player_stat(self, player_stat_map: Dict[str, PlayerStats], game: Game, proto_game: game_pb2.Game,
                              player_map: Dict[str, Player], data_frame: pd.DataFrame):
        for kickoff in proto_game.game_stats.kickoff_stats:
            for kickoff_player in kickoff.touch.players:
                if kickoff_player.player.id in player_stat_map:
                    is_first_touch = kickoff.touch.first_touch_player.id == kickoff_player.player.id
                    self.compute_totals(kickoff_player, player_stat_map[kickoff_player.player.id].kickoff_stats, is_first_touch)

        for player_id in player_stat_map:
            kickoff_stats = player_stat_map[player_id].kickoff_stats
            kickoff_stats.average_boost_used = kickoff_stats.average_boost_used / kickoff_stats.total_kickoffs

    def compute_totals(self, kickoff_player: KickoffPlayer,
                       player_kickoff_stats: CumulativeKickoffStats,
                       is_first_touch: bool):
        player_kickoff_stats.total_kickoffs += 1
        if kickoff_player.touch_position == BOOST:
            player_kickoff_stats.num_time_boost += 1
        if kickoff_player.touch_position == CHEAT:
            player_kickoff_stats.num_time_cheat += 1
        if kickoff_player.touch_position == GOAL:
            player_kickoff_stats.num_time_defend += 1
        if kickoff_player.touch_position == BALL:
            player_kickoff_stats.num_time_go_to_ball += 1
        if kickoff_player.touch_position == AFK:
            player_kickoff_stats.num_time_afk += 1
        if is_first_touch:
            player_kickoff_stats.num_time_first_touch += 1

        player_kickoff_stats.average_boost_used += kickoff_player.boost



from carball.json_parser.game import Game
from carball.generated.api.metadata.game_metadata_pb2 import RANKED_DROPSHOT, UNRANKED_DROPSHOT


def is_dropshot(game: Game):
    if game is None or game.game_info is None:
        return False
    return game.game_info.playlist in [RANKED_DROPSHOT, UNRANKED_DROPSHOT] or game.map == 'ShatterShot_P'



import logging

import pandas as pd

from ...json_parser.game import Game
logger = logging.getLogger(__name__)


class SaltieGame:

    @staticmethod
    def get_first_touch_frames(game):
        """
        Returns the frames for which the ball was touched in a valid way (if any touches happened)
        """
        if game.frames.loc[:, 'ball_has_been_hit'].any():
            ball_has_been_hit = game.frames.loc[:, 'ball_has_been_hit']
            last_frame_ball_has_been_hit = ball_has_been_hit.shift(1).rename('last_frame_ball_has_been_hit')
            ball_hit_dataframe = pd.concat([ball_has_been_hit, last_frame_ball_has_been_hit], axis=1)
            ball_hit_dataframe.fillna(False, inplace=True)

            kickoff_frames = ball_hit_dataframe[(ball_hit_dataframe['ball_has_been_hit']) &
                                                ~(ball_hit_dataframe['last_frame_ball_has_been_hit'])]
        else:
            logger.debug("No ball_has_been_hit?! Is this really old or what.")
            hit_team_no = game.ball.loc[:, 'hit_team_no']
            last_hit_team_no = hit_team_no.shift(1).rename('last_hit_team_no')
            hit_team_no_dataframe = pd.concat([hit_team_no, last_hit_team_no], axis=1)
            kickoff_frames = hit_team_no_dataframe[~(hit_team_no_dataframe['hit_team_no'].isnull()) &
                                                   (hit_team_no_dataframe['last_hit_team_no'].isnull())]

        return kickoff_frames.index.values

    @staticmethod
    def get_kickoff_frames(game):
        """
        Returns the frames between the kickoff countdown and the first touch occurring (if any touches happened).
        """
        if game.frames.loc[:, 'replicated_seconds_remaining'].any():
            ball_has_been_hit = game.frames.loc[:, 'replicated_seconds_remaining']
            last_frame_ball_has_been_hit = ball_has_been_hit.shift(1).rename('last_replicated_seconds_remaining')
            ball_hit_dataframe = pd.concat([ball_has_been_hit, last_frame_ball_has_been_hit], axis=1)
            ball_hit_dataframe.fillna(-1, inplace=True)

            countdown_start_frames = ball_hit_dataframe[(ball_hit_dataframe['replicated_seconds_remaining'] > 0) &
                                                  (ball_hit_dataframe['last_replicated_seconds_remaining'] == -1)]

            kickoff_frames = ball_hit_dataframe[(ball_hit_dataframe['replicated_seconds_remaining'] == 0) &
                                                (ball_hit_dataframe['last_replicated_seconds_remaining'] > 0)]

            countdown_indexes = countdown_start_frames.index.values

            if len(kickoff_frames.index.values) == len(countdown_indexes) == 1:
                reset_kickoff_index = kickoff_frames.index.values
            else:
                kickoff_count = 0
                reset_kickoff_index = []
                current_kickoff = None
                for kickoff in kickoff_frames.index.values:
                    if not kickoff_count + 1 < len(countdown_indexes):
                        break
                    if kickoff > countdown_indexes[kickoff_count] and kickoff < countdown_indexes[kickoff_count + 1]:
                        current_kickoff = kickoff
                    else:
                        reset_kickoff_index.append(current_kickoff)
                        kickoff_count += 1
                        current_kickoff = kickoff

                if current_kickoff is not None:
                    reset_kickoff_index.append(current_kickoff)

        else:
            logger.debug("No ball_has_been_hit?! Is this really old or what.")
            hit_team_no = game.ball.loc[:, 'hit_team_no']
            last_hit_team_no = hit_team_no.shift(1).rename('last_hit_team_no')
            hit_team_no_dataframe = pd.concat([hit_team_no, last_hit_team_no], axis=1)
            kickoff_frames = hit_team_no_dataframe[~(hit_team_no_dataframe['hit_team_no'].isnull()) &
                                                   (hit_team_no_dataframe['last_hit_team_no'].isnull())]

            reset_kickoff_index = kickoff_frames

        return reset_kickoff_index

    @staticmethod
    def create_data_df(game: Game) -> pd.DataFrame:
        data_dict = {player.name: player.data for player in game.players}
        data_dict['ball'] = game.ball
        initial_df = pd.concat(data_dict, axis=1)

        data_frame = pd.concat([initial_df, game.frames], axis=1)
        cols = []
        for c in data_frame.columns.values:
            if isinstance(c, str):
                cols.append(('game', c))
            else:
                cols.append(c)
        data_frame.columns = pd.MultiIndex.from_tuples(cols)
        return data_frame



from carball.analysis.analysis_manager import AnalysisManager
from carball.tests.utils import run_analysis_test_on_replay, get_raw_replays


class Test_Leader():

    def test_0_play_station_only_party(self, replay_cache):
        def test(analysis: AnalysisManager):
            game = analysis.game
            ps_party_leader = "6083491126912347959"
            ps_team_member_01 = "4674819165248336722"
            ps_team_member_03 = "4911491436059516465"
            ps_team_member_02 = ps_party_leader

            assert(ps_party_leader in game.parties.keys())
            assert(ps_team_member_01 in game.parties[ps_party_leader])
            assert(ps_team_member_02 in game.parties[ps_party_leader])
            assert(ps_team_member_03 in game.parties[ps_party_leader])

        run_analysis_test_on_replay(test, get_raw_replays()["PLAY_STATION_ONLY_PARTY"], cache=replay_cache)

    def test_1_party_leader_system_id_0(self, replay_cache):
        def test(analysis: AnalysisManager):
            game = analysis.game

            party_01_leader = "76561198084378722"
            party_01_member_01 = party_01_leader
            party_01_member_02 = "2904386747031141117"

            party_02_leader = "5948494783184915748"
            party_02_member_01 = party_02_leader
            party_02_member_02 = "1850445886414578837"

            assert (party_01_leader in game.parties.keys())
            assert (party_01_member_01 in game.parties[party_01_leader])
            assert (party_01_member_02 in game.parties[party_01_leader])

            assert (party_02_leader in game.parties.keys())
            assert (party_02_member_01 in game.parties[party_02_leader])
            assert (party_02_member_02 in game.parties[party_02_leader])

        run_analysis_test_on_replay(test, get_raw_replays()["PARTY_LEADER_SYSTEM_ID_0_ERROR"], cache=replay_cache)

    def test_2_xbox_party(self, replay_cache):
        def test(analysis: AnalysisManager):
            game = analysis.game

            party_01_leader = "76561198084378722"
            party_01_member_01 = party_01_leader
            party_01_member_02 = "2904386747031141117"

            party_02_leader = "2535465181947426"
            party_02_member_01 = party_02_leader
            party_02_member_02 = "2535416417939826"

            assert(party_01_leader in game.parties.keys())
            assert(party_01_member_01 in game.parties[party_01_leader])
            assert(party_01_member_02 in game.parties[party_01_leader])

            assert(party_02_leader in game.parties.keys())
            assert(party_02_member_01 in game.parties[party_02_leader])
            assert(party_02_member_02 in game.parties[party_02_leader])

        run_analysis_test_on_replay(test, get_raw_replays()["XBOX_PARTY"], cache=replay_cache)



from collections import Counter, defaultdict
from typing import List, Dict, Tuple

import pandas as pd

from carball.generated.api.stats.team_stats_pb2 import TeamStats
from ....analysis.stats.stats import BaseStat
from ....generated.api import game_pb2
from ....generated.api.player_pb2 import Player
from ....generated.api.stats.events_pb2 import Hit
from ....generated.api.stats.per_possession_stats_pb2 import PerPossessionStats
from ....generated.api.stats.player_stats_pb2 import PlayerStats
from ....json_parser.game import Game


class PerPossessionStat(BaseStat):
    possession_hit_fields = [
        'pass_', 'passed',
        'dribble', 'dribble_continuation',
        'shot', 'goal',
        'assist', 'assisted',
        'save',
        'aerial',
    ]

    def __init__(self):
        super().__init__()
        self.initialized = False

        self.player_possessions = []
        self.team_possessions = []

        self.player_possessions_dict = {}
        self.team_possessions_dict = {}

    def initialize(self, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                   data_frame: pd.DataFrame):
        _player_possessions, _team_possessions = self.get_possessions(proto_game, player_map)

        for possessions in (_player_possessions, _team_possessions):
            self.add_possession_durations(possessions, data_frame)

        # Remove possessions that are 1-hit or <1s
        self.player_possessions = [
            player_possession for player_possession in _player_possessions
            if len(player_possession.hits) > 1 and player_possession.duration > 1
        ]
        self.team_possessions = [
            team_possession for team_possession in _team_possessions
            if len(team_possession.hits) > 1 and team_possession.duration > 1
        ]
        self.logger.debug(f"Found {len(self.player_possessions)} player possessions and "
                         f"{len(self.team_possessions)} team possessions.")

        for possessions in (self.player_possessions, self.team_possessions):
            self.add_possession_counts(possessions)
            self.add_possession_distances(possessions)

        # Sort player_possessions by player
        self.player_possessions_dict = defaultdict(list)
        for possession in self.player_possessions:
            self.player_possessions_dict[possession.player_id].append(possession)

        # Sort team_possessions by team
        self.team_possessions_dict: Dict[bool, List['TeamPossession']] = defaultdict(list)
        for possession in self.team_possessions:
            self.team_possessions_dict[possession.is_orange].append(possession)

        self.initialized = True

    def calculate_player_stat(self, player_stat_map: Dict[str, PlayerStats], game: Game, proto_game: game_pb2.Game,
                              player_map: Dict[str, Player], data_frame: pd.DataFrame):
        if not self.initialized:
            self.initialize(proto_game, player_map, data_frame)

        player_possession_stats_dict: Dict[str, List['PlayerPossession']] = {}
        for player_id, possessions in self.player_possessions_dict.items():
            stats = self.get_per_possession_stats(possessions)

            player_stat_map[player_id].per_possession_stats.CopyFrom(stats)

            player_possession_stats_dict[player_id] = stats

        for player_id, player_stats in player_stat_map.items():
            if player_id in player_possession_stats_dict:
                player_stats.per_possession_stats.CopyFrom(player_possession_stats_dict[player_id])

    def calculate_team_stat(self, team_stat_list: Dict[int, TeamStats], game: Game, proto_game: game_pb2.Game,
                            player_map: Dict[str, Player], data_frame: pd.DataFrame):
        for team_is_orange, possessions in self.team_possessions_dict.items():
            stats = self.get_per_possession_stats(possessions)
            team_stat_list[team_is_orange].per_possession_stats.CopyFrom(stats)

    @staticmethod
    def get_possessions(proto_game: game_pb2.Game,
                        player_map: Dict[str, Player]) -> Tuple[List['PlayerPossession'], List['TeamPossession']]:
        hits = proto_game.game_stats.hits

        player_possessions: List[PlayerPossession] = []
        team_possessions: List[TeamPossession] = []

        current_goal_number = 0

        current_player_possession = None
        current_team_possession = None

        for hit in hits:
            player_id = hit.player_id.id
            player = player_map[player_id]

            if current_goal_number != hit.goal_number:
                # End current possessions before next goal
                if current_player_possession is not None:
                    player_possessions.append(current_player_possession)
                    current_player_possession = None
                if current_team_possession is not None:
                    team_possessions.append(current_team_possession)
                    current_team_possession = None

            current_goal_number = hit.goal_number

            if current_team_possession is None:
                # Start possession
                current_team_possession = TeamPossession([hit], player.is_orange)
            else:
                if current_team_possession.is_orange == player.is_orange:
                    # Add to current_team_possession
                    current_team_possession.hits.append(hit)
                else:
                    # Record last possession and start new
                    team_possessions.append(current_team_possession)
                    current_team_possession = TeamPossession([hit], player.is_orange)

            if current_player_possession is None:
                # Start possession
                current_player_possession = PlayerPossession([hit], player_id)
            else:
                if current_player_possession.player_id == player_id:
                    # Add to current_player_possession
                    current_player_possession.hits.append(hit)
                else:
                    # Record last possession and start new
                    player_possessions.append(current_player_possession)
                    current_player_possession = PlayerPossession([hit], player_id)

        return player_possessions, team_possessions

    @staticmethod
    def add_possession_durations(possessions: List['BasePossession'], data_frame: pd.DataFrame):
        player_possessions_count = len(possessions)
        for i, possession in enumerate(possessions):
            first_hit = possession.hits[0]
            possession_start_frame = first_hit.frame_number

            if i != player_possessions_count - 1:
                # Next possession exists
                next_possession = possessions[i + 1]
                if next_possession.hits[0].goal_number != first_hit.goal_number:
                    # Next possession is in next goal frame
                    # possession end frame is end of goal frame
                    possession_end_frame = \
                        data_frame.index[(data_frame.loc[:, ('game', 'goal_number')] == first_hit.goal_number)][-1]
                else:
                    # possession end frame is first hit of next possession
                    possession_end_frame = next_possession.hits[0].frame_number
            else:
                # Last possession of game
                # possession end frame is last frame of game
                possession_end_frame = data_frame.index.max()

            try:
                possession_times = data_frame.loc[[possession_start_frame, possession_end_frame], ('game', 'time')].values
                possession_duration = possession_times[1] - possession_times[0]
            except KeyError:
                possession_duration = 0
            possession.duration = possession_duration

    @classmethod
    def add_possession_counts(cls, possessions: List['BasePossession']):
        for possession in possessions:
            for hit in possession.hits:
                for field in cls.possession_hit_fields:
                    if getattr(hit, field):
                        possession.counts[field] += 1

    @staticmethod
    def add_possession_distances(possessions: List['BasePossession']):
        for possession in possessions:
            distance = 0
            for hit in possession.hits:
                distance += hit.distance
            possession.distance = distance

    @classmethod
    def get_per_possession_stats(cls, possessions: List['BasePossession']) -> PerPossessionStats:
        count = len(possessions)
        hit_counts = [len(possession.hits) for possession in possessions]
        average_hits = sum(hit_counts) / len(hit_counts)
        durations = [possession.duration for possession in possessions]
        average_duration = sum(durations) / len(durations)

        average_counts = {}

        for field in cls.possession_hit_fields:
            sum_ = sum([possession.counts[field] for possession in possessions])
            average_counts[field] = sum_ / count

        per_possession_stats = PerPossessionStats()
        cls.add_average_counts(per_possession_stats, average_counts)

        per_possession_stats.average_hits = average_hits
        per_possession_stats.average_duration = average_duration
        per_possession_stats.count = count

        return per_possession_stats

    @staticmethod
    def add_average_counts(per_possession_stats: PerPossessionStats, average_counts: Dict[str, float]):
        for field, value in average_counts.items():
            setattr(per_possession_stats.average_counts, field, value)


class BasePossession:
    def __init__(self, hits: List[Hit]):
        self.hits = hits

        # Added later
        self.duration = None
        self.distance = None
        self.counts = Counter()


class PlayerPossession(BasePossession):
    def __init__(self, hits: List[Hit], player_id: str):
        super().__init__(hits)
        self.player_id = player_id


class TeamPossession(BasePossession):
    def __init__(self, hits: List[Hit], is_orange: bool):
        super().__init__(hits)
        self.is_orange = is_orange



from .base import *


class CameraSettingsHandler(BaseActorHandler):
    type_name = 'TAGame.Default__CameraSettingsActor_TA'

    def update(self, actor: dict, frame_number: int, time: float, delta: float) -> None:
        if 'TAGame.CameraSettingsActor_TA:PRI' not in actor:
            return

        player_actor_id = actor['TAGame.CameraSettingsActor_TA:PRI']['actor']  # may need to try another key
        # add camera settings
        if player_actor_id not in self.parser.cameras_data and \
                'TAGame.CameraSettingsActor_TA:ProfileSettings' in actor:
            self.parser.cameras_data[player_actor_id] = actor['TAGame.CameraSettingsActor_TA:ProfileSettings']
        # add ball cam to inputs
        ball_cam = actor.get('TAGame.CameraSettingsActor_TA:bUsingSecondaryCamera', None)
        try:
            self.parser.player_data[player_actor_id][frame_number]['ball_cam'] = ball_cam
        except KeyError:
            # key error due to frame_number not in inputs
            # ignore as no point knowing
            pass



import pytest

from carball.command_line import main
from carball.tests.utils import get_replay_path, get_raw_replays


class Test_Commandline():
    def test_command_line_valid_input(self):
        replay_path = get_replay_path(get_raw_replays()["KICKOFF_NO_TOUCH"][0])
        main(program_args=["--i", replay_path, "-v", "--dry-run"])

    def test_command_line_invalid_input(self):
        replay_path = get_replay_path("INVALID_REPLAY")
        with pytest.raises(FileNotFoundError):
            main(program_args=["--i", replay_path, "-s", "--proto", "PATH_DOESNT_MATTER"])

    def test_command_line_no_output(self):
        replay_path = get_replay_path(get_raw_replays()["KICKOFF_NO_TOUCH"][0])
        with pytest.raises(SystemExit):
            main(program_args=["--i", replay_path])



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/party.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..api import player_id_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/party.proto',
  package='api',
  serialized_pb=_b('\n\x0f\x61pi/party.proto\x12\x03\x61pi\x1a\x13\x61pi/player_id.proto\"I\n\x05Party\x12 \n\tleader_id\x18\x01 \x01(\x0b\x32\r.api.PlayerId\x12\x1e\n\x07members\x18\x02 \x03(\x0b\x32\r.api.PlayerId')
  ,
  dependencies=[player_id_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_PARTY = _descriptor.Descriptor(
  name='Party',
  full_name='api.Party',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='leader_id', full_name='api.Party.leader_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='members', full_name='api.Party.members', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=45,
  serialized_end=118,
)

_PARTY.fields_by_name['leader_id'].message_type = player_id_pb2._PLAYERID
_PARTY.fields_by_name['members'].message_type = player_id_pb2._PLAYERID
DESCRIPTOR.message_types_by_name['Party'] = _PARTY

Party = _reflection.GeneratedProtocolMessageType('Party', (_message.Message,), dict(
  DESCRIPTOR = _PARTY,
  __module__ = 'api.party_pb2'
  # @@protoc_insertion_point(class_scope:api.Party)
  ))
_sym_db.RegisterMessage(Party)


# @@protoc_insertion_point(module_scope)



from .base import *


class ActiveHandler(BaseActorHandler):

    def __init__(self, parser: object, data_key: str):
        super().__init__(parser)
        self.data_key = data_key

    def update(self, actor: dict, frame_number: int, time: float, delta: float) -> None:
        car_actor_id = actor.get('TAGame.CarComponent_TA:Vehicle', {}).get('actor', None)
        if car_actor_id is not None:
            if car_actor_id in self.parser.current_car_ids_to_collect:
                player_actor_id = self.parser.car_player_ids[car_actor_id]
                jump_is_active = actor.get(
                    COMPONENT_ACTIVE_KEY,
                    actor.get(COMPONENT_REPLICATED_ACTIVE_KEY, False))
                self.parser.player_data[player_actor_id][frame_number][self.data_key] = jump_is_active


class JumpHandler(ActiveHandler):
    type_name = 'Archetypes.CarComponents.CarComponent_Jump'

    def __init__(self, parser: object):
        super().__init__(parser, 'jump_active')


class DoubleJumpHandler(ActiveHandler):
    type_name = 'Archetypes.CarComponents.CarComponent_DoubleJump'

    def __init__(self, parser: object):
        super().__init__(parser, 'double_jump_active')


class DodgeHandler(ActiveHandler):
    type_name = 'Archetypes.CarComponents.CarComponent_Dodge'

    def __init__(self, parser: object):
        super().__init__(parser, 'dodge_active')



import os
import time
from typing import Callable, Tuple, Iterable

import numpy as np
from carball.decompile_replays import analyze_replay_file

REPLAYS_FOLDER = os.path.join(os.path.dirname(__file__), 'replays')


def get_multiple_answers(answers: Iterable[str]) -> Iterable[Tuple[any]]:
    return np.column_stack([get_specific_answers().get(key) for key in answers])


def run_tests_on_list(unit_test_func: Callable, replay_list=None, answers=None):
    if replay_list is None:
        replay_list = get_complex_replay_list()

    for index in range(len(replay_list)):
        replay_file = replay_list[index]
        print('running test on replay: ' + replay_file)
        answer = answers[index] if answers is not None and index < len(answers) else None
        run_replay(replay_file, unit_test_func, answer=answer)


def get_replay_path(replay_file):
    return os.path.join(REPLAYS_FOLDER, replay_file)


def run_replay(replay_file, unit_test_func: Callable, answer=None):
    """
    Runs the replay with the file downloaded locally then deletes the file.
    :param replay_file:
    :param unit_test_func:
    :param answer: data that can be passed to the replay to help judge it
    :return:
    """
    file = get_replay_path(replay_file)

    if answer is not None:
        unit_test_func(file, answer)
    else:
        unit_test_func(file)


def run_analysis_test_on_replay(unit_test_func: Callable, replay_list=None, answers=None, cache=None, calculate_intensive_events=False):
    """
    :param unit_test_func: Called with an AnalysisManager
    :param replay_list: list of replay urls
    :param calculate_intensive_events: Indicates if the test needs the replay to be analyzed against expensive-to-calculate events.
    :return:
    """

    def wrapper(replay_file_path, answer=None):
        start = time.time()
        if cache is not None and str(replay_file_path) in cache:
            analysis_manager = cache[str(replay_file_path)]
        else:
            analysis_manager = analyze_replay_file(replay_file_path, calculate_intensive_events=calculate_intensive_events)
        if cache is not None and time.time() - start > 10:
            cache[str(replay_file_path)] = analysis_manager
        if answer is not None:
            if isinstance(answer, (list, tuple, np.ndarray)):
                unit_test_func(analysis_manager, *answer)
            else:
                unit_test_func(analysis_manager, answer)
        else:
            unit_test_func(analysis_manager)

    run_tests_on_list(wrapper, replay_list, answers=answers)


def get_complex_replay_list():
    """
    For full replays that have crashed or failed to be converted
    :return:
    """
    return [
        'BOTS_JOINING_AND_LEAVING.replay',
        'BOTS_NO_POSITION.replay',
        'ZEROED_STATS.replay',
        'FAKE_BOTS_SkyBot.replay',
        'NEGATIVE_WASTED_COLLECTION.replay',
        'WASTED_BOOST_WHILE_SUPER_SONIC.replay',
        'OCE_RLCS_7_CARS.replay',
        'crossplatform_party.replay',
        'PARTY_LEADER_SYSTEM_ID_0.replay',
        'RL_1_66.replay',
        'CARRY.replay'
    ]


def get_raw_replays():
    """
    Partial replays for specific values.
    :return:
    """
    return {
        "0_JUMPS": ["0_JUMPS.replay"],
        "0_SAVES": ["0_SAVES.replay"],
        "1_AERIAL": ["1_AERIAL.replay"],
        "1_DEMO": ["1_DEMO.replay", "1_DEMO_1_63.replay"],
        "1_DOUBLE_JUMP": ["1_DOUBLE_JUMP.replay"],
        "1_EPIC_SAVE": ["1_EPIC_SAVE.replay"],
        "1_JUMP": ["1_JUMP.replay"],
        "1_NORMAL_SAVE_FROM_SHOT_TOWARD_POST": ["1_NORMAL_SAVE.replay"],

        # Boost
        "3_STEAL_ORANGE_0_STEAL_BLUE": ["3_STEALS.replay"],
        "12_BOOST_PAD_0_USED": ["12_BOOST_PAD_0_USED.replay"],
        "12_BOOST_PAD_45_USED": ["12_BOOST_PAD_45_USED.replay"],
        "100_BOOST_PAD_0_USED": ["100_BOOST_PAD_0_USED.replay"],
        "100_BOOST_PAD_100_USED": ["100_BOOST_PAD_100_USED.replay"],
        "NO_BOOST_PAD_0_USED": ["NO_BOOST_PAD_0_USED.replay"],
        "NO_BOOST_PAD_33_USED": ["NO_BOOST_PAD_33_USED.replay"],
        "12_AND_100_BOOST_PADS_0_USED": ["12_AND_100_BOOST_PADS_0_USED.replay", "1_100_1_12_0_BOOST_168.replay", "1_100_1_12_0_BOOST_168_ALT.replay"],
        "WASTED_BOOST_WHILE_SUPER_SONIC": ["WASTED_BOOST_WHILE_SUPER_SONIC.replay"],
        "CALCULATE_USED_BOOST_WITH_DEMO": ["CALCULATE_USED_BOOST_WITH_DEMO.replay"],
        "CALCULATE_USED_BOOST_DEMO_WITH_FLIPS": ["CALCULATE_USED_BOOST_DEMO_WITH_FLIPS.replay"],
        "MORE_THAN_100_BOOST": ["MORE_THAN_100_BOOST.replay"],
        "USE_BOOST_AFTER_GOAL": ["USE_BOOST_AFTER_GOAL.replay"],
        "FEATHERING_34x100_BO0ST_USED": ["FEATHERING_34_X_100_BOOSTS_USED.replay"],
        "6_BIG_25_SMALL": ["6_BIG_25_SMALL.replay"],

        # Kickoffs
        "STRAIGHT_KICKOFF_GOAL": ["Straight_Kickoff_Goal.replay"],
        "KICKOFF_NO_TOUCH": ["NO_KICKOFF.replay"],
        "LAST_KICKOFF_NO_TOUCH": ["LAST_KICKOFF_NO_TOUCH.replay"],
        "3_KICKOFFS": ["3_KICKOFFS_4_SHOTS.replay"],
        "5_DIVERSE_KICKOFFS": ["KICKOFF_TEST_1.replay"],
        "6_DIVERSE_KICKOFFS": ["KICKOFF_TEST_2.replay"],
        "NO_ONE_FOR_KICKOFF": ["NO_ONE_FOR_KICKOFF.replay"],
        "WEIRD_KICKOFFS": ["WEIRD_KICKOFFS.replay"],

        # hits
        "4_SHOTS": ["3_KICKOFFS_4_SHOTS.replay"],
        "KICKOFF_3_HITS": ["KICKOFF_3_HITS.replay"],
        "MID_AIR_PASS": ["MID_AIR_PASS_GOAL.replay"],
        "HIGH_AIR_PASS": ["HIGH_AIR_PASS_GOAL.replay"],
        "GROUND_PASS": ["GROUNDED_PASS_GOAL.replay"],
        "PINCH_GROUND": ["PINCH_GROUNDED_GOAL.replay"],
        "DEFAULT_3_ON_3_AROUND_58_HITS": ["DEFAULT_3_ON_3_AROUND_58_HITS.replay"],
        "1_CLEAR": ["1_CLEAR.replay"],
        "2_CLEARS": ["2_CLEARS.replay"],

        # KBM
        "1_MIN_KBM_1_MIN_XBO_CONTROLLER": ["1_MIN_KBM_1_MIN_XBO_CONTROLLER.replay"],
        "100_PERCENT_KBM": ["100_PERCENT_KBM.replay"],

        # camera controls
        "BALLCAM_ON_IN_CLOSE_AIR_ELSE_OFF": ["BALLCAM_ON_IN_CLOSE_AIR_ELSE_OFF.replay"],
        "BALLCAM_OFF_IN_CLOSE_AIR_ELSE_ON": ["BALLCAM_OFF_IN_CLOSE_AIR_ELSE_ON.replay"],
        "BALLCAM_MIXED": ["BALLCAM_MIXED.replay"],
        "BALLCAM_ON_AT_DRIBBLE_ELSE_OFF": ["BALLCAM_ON_AT_DRIBBLE_ELSE_OFF.replay"],
        "BALLCAM_OFF_AT_DRIBBLE_ELSE_ON": ["BALLCAM_OFF_AT_DRIBBLE_ELSE_ON.replay"],
        "BALLCAM_OFF_ALWAYS": ["BALLCAM_OFF_ALWAYS.replay"],
        "BALLCAM_ON_ALWAYS": ["BALLCAM_ON_ALWAYS.replay"],

        # Dribbles
        "SKYBOT_DRIBBLE_INFO": ["SKYBOT_DRIBBLE_INFO.replay"],
        "1_DRIBBLE": ["1_DRIBBLE.replay"],
        "3_DRIBBLE_2_FLICKS": ["3_DRIBBLE_2_FLICKS.replay"],

        # parties
        "PLAY_STATION_ONLY_PARTY": ['PLAY_STATION_ONLY_PARTY.replay'],
        # error cases
        "PLAYERNAME_BALL": ["PLAYERNAME_BALL.replay"],
        "PLAYERNAME_GAME": ["PLAYERNAME_GAME.replay"],
        "PLAYERNAME_TWO_PLAYERS_NAMED_SAME": ["PLAYERNAME_TWO_PLAYERS_NAMED_SAME.replay"],
        "PLAYERNAME_ZTTL": ["PLAYERNAME_ZTTL.replay"],
        "ISSUE_PLAYER_REJOIN": ["ISSUE_PLAYER_REJOIN.replay"],
        "OCE_RLCS_7_CARS": ["OCE_RLCS_7_CARS.replay"],
        "XBOX_PARTY": ["XBOX_PARTY.replay"],
        # error cases
        "UNICODE_ERROR": ["UnicodeEncodeError.replay"],
        "CROSSPLATFORM_PARTY_LEADER_ERROR": ["crossplatform_party.replay"],
        "PARTY_LEADER_SYSTEM_ID_0_ERROR": ['PARTY_LEADER_SYSTEM_ID_0.replay'],

        # rumble
        "RUMBLE_PRE_ITEM_GOALS": ["RUMBLE_PRE_ITEM_GOALS.replay"],
        "RUMBLE_ITEM_GOALS": ["RUMBLE_ITEM_GOALS.replay"],
        "RUMBLE_FREEZE_VS_SPIKE": ["RUMBLE_FREEZE_VS_SPIKE.replay"],
        "RUMBLE_HOLD_TIME": ["RUMBLE_HOLD_TIME.replay"],
        "RUMBLE_FULL": ["RUMBLE_FULL.replay"],
        "RUMBLE_ITEM_KICKOFF": ["RUMBLE_ITEM_KICKOFF.replay"],

        # dropshot
        "DROPSHOT_SINGLE_DAMAGE": ["DROPSHOT_SINGLE_DAMAGE.replay"],
        "DROPSHOT_DOUBLE_DAMAGE": ["DROPSHOT_DOUBLE_DAMAGE.replay"],
        "DROPSHOT_PHASE1_BALL": ["DROPSHOT_PHASE1_BALL.replay"],
        "DROPSHOT_PHASE2_BALL": ["DROPSHOT_PHASE2_BALL.replay"],
        "DROPSHOT_GOAL": ["DROPSHOT_GOAL.replay"],

        # ratteltrap errors
        "BROKEN_REPLAY": ["INVALID_FILE.replay"]
    }


def get_specific_replays():
    raw_map = get_raw_replays()
    return {
        # BOOSTS
        "0_BOOST_COLLECTED": raw_map["NO_BOOST_PAD_0_USED"] + raw_map["NO_BOOST_PAD_33_USED"] +
                             raw_map["KICKOFF_NO_TOUCH"],
        "1_SMALL_PAD": raw_map["12_BOOST_PAD_0_USED"] + raw_map["12_BOOST_PAD_45_USED"] +
                        raw_map['12_AND_100_BOOST_PADS_0_USED'],
        "1_LARGE_PAD": raw_map["100_BOOST_PAD_0_USED"] + raw_map["100_BOOST_PAD_100_USED"] +
                        raw_map['12_AND_100_BOOST_PADS_0_USED'],
        "0_BOOST_USED": raw_map["12_BOOST_PAD_0_USED"] + raw_map["100_BOOST_PAD_0_USED"] +
                        raw_map["NO_BOOST_PAD_0_USED"] + raw_map["KICKOFF_NO_TOUCH"] +
                        raw_map['12_AND_100_BOOST_PADS_0_USED'],
        "BOOST_USED": raw_map["12_BOOST_PAD_45_USED"] +
                      raw_map["100_BOOST_PAD_100_USED"] +
                      raw_map["NO_BOOST_PAD_33_USED"] +
                      raw_map["CALCULATE_USED_BOOST_WITH_DEMO"] +
                      raw_map["CALCULATE_USED_BOOST_DEMO_WITH_FLIPS"] +
                      raw_map["USE_BOOST_AFTER_GOAL"] +
                      raw_map["WASTED_BOOST_WHILE_SUPER_SONIC"],
        "BOOST_FEATHERED": raw_map["MORE_THAN_100_BOOST"] + raw_map["FEATHERING_34x100_BO0ST_USED"],
        "BOOST_WASTED_USAGE": raw_map["WASTED_BOOST_WHILE_SUPER_SONIC"],
        "BOOST_WASTED_COLLECTION": raw_map["MORE_THAN_100_BOOST"] + raw_map["12_AND_100_BOOST_PADS_0_USED"],
        # HITS
        "HITS": raw_map["4_SHOTS"] + raw_map["KICKOFF_3_HITS"] + raw_map["12_BOOST_PAD_45_USED"] +
                raw_map["MID_AIR_PASS"] + raw_map["HIGH_AIR_PASS"] + raw_map["GROUND_PASS"] +
                raw_map["1_NORMAL_SAVE_FROM_SHOT_TOWARD_POST"] + raw_map["1_EPIC_SAVE"] + raw_map["1_AERIAL"] +
                raw_map["DEFAULT_3_ON_3_AROUND_58_HITS"],
        # + raw_map["PINCH_GROUND"],  TODO: Fix pinches to create 2 hits 1 for each person on same frame
        "SHOTS": raw_map["4_SHOTS"] + raw_map["12_BOOST_PAD_45_USED"] +
                 raw_map["1_EPIC_SAVE"] + raw_map["1_NORMAL_SAVE_FROM_SHOT_TOWARD_POST"],
        "PASSES": raw_map["MID_AIR_PASS"] + raw_map["HIGH_AIR_PASS"] + raw_map["GROUND_PASS"],
        "AERIALS": raw_map["1_EPIC_SAVE"] + raw_map["1_AERIAL"] + raw_map["HIGH_AIR_PASS"] + raw_map["MID_AIR_PASS"],
        "CANT_CRASH": raw_map["UNICODE_ERROR"] + raw_map["PLAYERNAME_BALL"] + raw_map["PLAYERNAME_GAME"] +
                      raw_map["PLAYERNAME_TWO_PLAYERS_NAMED_SAME"] + raw_map["PLAYERNAME_ZTTL"] +
                      raw_map["ISSUE_PLAYER_REJOIN"] + raw_map["ISSUE_PLAYER_REJOIN"],
        "ZERO_DRIBBLE": raw_map["12_BOOST_PAD_45_USED"] + raw_map["KICKOFF_NO_TOUCH"],
        "DRIBBLES": raw_map["1_DRIBBLE"] + raw_map["3_DRIBBLE_2_FLICKS"] + raw_map["SKYBOT_DRIBBLE_INFO"],
        "SAVES": raw_map["1_EPIC_SAVE"] + raw_map["1_NORMAL_SAVE_FROM_SHOT_TOWARD_POST"],
        "OFFLINE": raw_map["3_KICKOFFS"],
        "BROKEN_REPLAYS": raw_map["BROKEN_REPLAY"],
        "CLEARS": raw_map['1_CLEAR'] + raw_map['2_CLEARS']
    }


def get_specific_answers():
    specific_replays = get_specific_replays()
    return {
        # Boost
        "0_BOOST_USED": [0] * len(specific_replays["0_BOOST_USED"]),
        "BOOST_USED": [45, 100, 33, 33.33 + 33.33 + 12.15, 33.33, 33.33, 0],
        "BOOST_WASTED_USAGE": [33.33],
        "BOOST_WASTED_COLLECTION": [[((31, 238), (255, 85))], [([255, 116],)], [([255, 116],)], [([255, 116],)]],
        "BOOST_FEATHERED": [100.0, 3490.0],
        # Hits
        "HITS": [4, 3, 1, 2, 9, 2, 4, 4, 4, 50],
        "SHOTS": [3, 0, 2, 1],
        "PASSES": [1, 1, 1],
        "AERIALS": [0, 1, 2, 0],
        "DRIBBLES": [1, 3, 49],
        "FLICKS": [0, 1, 0],
        "SAVES": [1, 0],
        "CLEARS": [1, 2],
    }


def assertNearlyEqual(case, a, b, percent=2.0, msg=None):
    if abs(a - b) > abs(percent / 100.0 * min(abs(a), abs(b))):
        if msg is None:
            case.fail("The given numbers %s and %s are not within %s percent of each other." % (a, b, percent))
        else:
            case.fail(msg)



from .actor import *

REPLICATED_RB_STATE_KEY = 'TAGame.RBActor_TA:ReplicatedRBState'

_HANDLERS = [
    GameInfoHandler,
    GameEventHandler,
    BallHandler,
    PlayerHandler,
    TeamHandler,
    CarHandler,
    JumpHandler,
    DodgeHandler,
    DoubleJumpHandler,
    BoostHandler,
    BoostPickupHandler,
    CameraSettingsHandler,
    RumbleItemHandler,
    PlatformHandler
]

_PRIORITY_HANDLERS = [
    GameInfoHandler,
    PlayerHandler,
    CarHandler
]

# These handlers will also handle frames with a delta of 0 to match the old implementation
# TODO check if this is needed
_0_DELTA_HANDLERS = [
    GameInfoHandler,
    PlayerHandler
]


def parse_frames(game):
    """
    :return: all_data = {
        'player_ball_data': player_ball_data,
        'player_dicts': player_dicts,
        'team_dicts': team_dicts,
        'frames_data': frames_data,
        'cameras_data': cameras_data
        'demos_data': demos_data
    }

    player_ball_data format:
    {
    'ball': {frame_number: {pos_x, pos_y ...}, f_no2: {...} ...,}
    player_actor_id: {
        frame_number: {
            pos_x, pos_y ...,
            throttle, steer, ...,
            ping, ball_cam
        },
        f_no2: {...} ...,
    }

    currently implemented:
        inputs: posx, posy, posz, rotx, roty, rotz, vx, vy, vz, angvx, angy, angvz, boost_amt
        outputs: throttle, steer, handbrake, boost, jump, doublejump, dodge

    player_dicts  = {player_actor_id : {actor_data}, player_actor_id_2: {actor_data_2}}
    team_dicts = {team_actor_id: {actor_data, 'colour':'blue'/'orange', also includes name}
    frames_data = {frame_number: {time, delta, seconds_remaining, is_overtime, ball_has_been_hit}
    cameras_data = {player_actor_id: actor_data}
    demos_data = {frame_number: demolish_data}

    """
    parser = FrameParser(game.replay_data, game)
    parser.parse_frames()

    player_ball_data = parser.player_data
    player_ball_data['ball'] = parser.ball_data

    return {
        'player_ball_data': player_ball_data,
        'player_dicts': parser.player_dicts,
        'car_dicts': parser.car_dicts,
        'team_dicts': parser.team_dicts,
        'frames_data': parser.frames_data,
        'cameras_data': parser.cameras_data,
        'demos_data': parser.demos_data,
        'game_info_actor': parser.game_info_actor,
        'soccar_game_event_actor': parser.soccar_game_event_actor,
        'parties': parser.parties,
        'dropshot': parser.dropshot
    }


class FrameParser(object):

    def __init__(self, replay_frames, game):
        self.replay_frames = replay_frames
        self.game = game
        self.replay_version = game.replay_version
        self.objects = game.replay['objects']
        self.names = game.replay['names']
        self.class_indices = game.replay['class_indices']

        self.game_info_actor = None
        self.soccar_game_event_actor = None

        self.frames_data = {}
        self.ball_data = {}
        self.player_data = {}

        # dictionaries to contain data in frames
        self.parties = {}
        self.player_dicts = {}
        self.team_dicts = {}
        self.car_dicts = {}

        self.player_car_ids = {}  # player_actor_id: car_actor_id
        self.car_player_ids = {}  # car_actor_id: player_actor_id

        self.cameras_data = {}  # player_actor_id: actor_data
        self.demos_data = []  # frame_number: demolish_data

        # stores car_actor_ids to collect data for at each frame
        self.current_car_ids_to_collect = []

        self.actors = {}

        self.dropshot = {
            'tile_states': {},
            'damage_events': {},
            'ball_state': 0,
            'ball_events': [],
            'tile_frames': {}
        }

    def parse_frames(self):

        self.actors = {}
        handlers = [dict() for _ in range(len(_PRIORITY_HANDLERS) + 1)]
        handled_actors = set()

        current_goal_number = 0

        for i, frame in enumerate(self.replay_frames):
            time = frame['time']
            delta = frame['delta']

            for actor_id in frame['deleted_actors']:
                for handler_group in handlers:
                    handler_group.pop(actor_id, None)
                handled_actors.discard(actor_id)
                self.player_car_ids.pop(actor_id, None)
                self.car_player_ids.pop(actor_id, None)
                self.actors.pop(actor_id, None)

            for new_actor in frame['new_actors']:
                actor_id = new_actor['actor_id']
                object_name = self.objects[new_actor['object_id']]
                self.actors[actor_id] = {
                    'Id': actor_id,
                    'TypeName': object_name,
                    'ClassName': OBJECT_CLASSES.get(object_name, None),
                    'Name': self.names[new_actor['name_id']],
                }

                handler = next(filter(lambda handler_cls: handler_cls.can_handle(self.actors[actor_id]),
                                      _HANDLERS), None)

                if handler is not None:
                    try:
                        priority = _PRIORITY_HANDLERS.index(handler)
                    except ValueError:
                        priority = len(_PRIORITY_HANDLERS)

                    handlers[priority][actor_id] = handler(self), handler in _0_DELTA_HANDLERS
                    handled_actors.add(actor_id)

            for updated in frame['updated_actors']:
                actor_id = updated['actor_id']

                if actor_id not in self.actors:
                    continue

                actor = self.actors[actor_id]
                prop_name = self.objects[updated['object_id']]
                # update property
                actor[prop_name] = find_actual_value(updated['attribute'])

            # stop data collection after goal
            try:
                if i > self.game.goals[current_goal_number].frame_number:
                    # set all players to sleeping after goal
                    for car_actor_id in self.car_player_ids:
                        try:
                            car = self.actors[car_actor_id]
                            car[REPLICATED_RB_STATE_KEY]['Sleeping'] = True
                        except KeyError as e:
                            # Ignore the case where the car does not have a REPLICATED_RB_STATE_KEY
                            pass
                    current_goal_number += 1
            except IndexError:
                # after last goal.
                pass

            # apply the update handlers
            for handler_group in handlers:
                for actor_id, handler_tuple in handler_group.items():
                    handler = handler_tuple[0]

                    # skip 0 delta frames, except for these two handlers (matches old implementation)
                    if handler_tuple[1] or delta != 0:
                        handler.update(self.actors[actor_id], i, time, delta)

            self.current_car_ids_to_collect.clear()


def find_actual_value(attribute: dict):
    item = list(attribute.items())[0]
    attribute_type = item[0]
    attribute_value = item[1]

    if attribute_type == 'Flagged' or attribute_type == 'FlaggedByte':
        return attribute_value[1]
    return attribute_value


OBJECT_CLASSES = {
    "Archetypes.Ball.Ball_BasketBall_Mutator": "TAGame.Ball_TA",
    "Archetypes.Ball.Ball_Basketball": "TAGame.Ball_TA",
    "Archetypes.Ball.Ball_BasketBall": "TAGame.Ball_TA",
    "Archetypes.Ball.Ball_Beachball": "TAGame.Ball_TA",
    "Archetypes.Ball.Ball_Breakout": "TAGame.Ball_Breakout_TA",
    "Archetypes.Ball.Ball_Default": "TAGame.Ball_TA",
    "Archetypes.Ball.Ball_Trajectory": "TAGame.Ball_TA",
    "Archetypes.Ball.Ball_Haunted": "TAGame.Ball_Haunted_TA",
    "Archetypes.Ball.Ball_Puck": "TAGame.Ball_TA",
    "Archetypes.Ball.Ball_Anniversary": "TAGame.Ball_TA",
    "Archetypes.Ball.CubeBall": "TAGame.Ball_TA",
    "Archetypes.Ball.Ball_Training": "TAGame.Ball_TA",
    "Archetypes.Car.Car_Default": "TAGame.Car_TA",
    "Archetypes.CarComponents.CarComponent_Boost": "TAGame.CarComponent_Boost_TA",
    "Archetypes.CarComponents.CarComponent_Dodge": "TAGame.CarComponent_Dodge_TA",
    "Archetypes.CarComponents.CarComponent_DoubleJump": "TAGame.CarComponent_DoubleJump_TA",
    "Archetypes.CarComponents.CarComponent_FlipCar": "TAGame.CarComponent_FlipCar_TA",
    "Archetypes.CarComponents.CarComponent_Jump": "TAGame.CarComponent_Jump_TA",
    "Archetypes.GameEvent.GameEvent_Basketball": "TAGame.GameEvent_Soccar_TA",
    "Archetypes.GameEvent.GameEvent_BasketballPrivate": "TAGame.GameEvent_SoccarPrivate_TA",
    "Archetypes.GameEvent.GameEvent_BasketballSplitscreen": "TAGame.GameEvent_SoccarSplitscreen_TA",
    "Archetypes.GameEvent.GameEvent_Breakout": "TAGame.GameEvent_Soccar_TA",
    "Archetypes.GameEvent.GameEvent_Hockey": "TAGame.GameEvent_Soccar_TA",
    "Archetypes.GameEvent.GameEvent_HockeyPrivate": "TAGame.GameEvent_SoccarPrivate_TA",
    "Archetypes.GameEvent.GameEvent_HockeySplitscreen": "TAGame.GameEvent_SoccarSplitscreen_TA",
    "Archetypes.GameEvent.GameEvent_Items": "TAGame.GameEvent_Soccar_TA",
    "Archetypes.GameEvent.GameEvent_Season:CarArchetype": "TAGame.Car_TA",
    "Archetypes.GameEvent.GameEvent_Season": "TAGame.GameEvent_Season_TA",
    "Archetypes.GameEvent.GameEvent_Soccar": "TAGame.GameEvent_Soccar_TA",
    "Archetypes.GameEvent.GameEvent_SoccarLan": "TAGame.GameEvent_Soccar_TA",
    "Archetypes.GameEvent.GameEvent_SoccarPrivate": "TAGame.GameEvent_SoccarPrivate_TA",
    "Archetypes.GameEvent.GameEvent_SoccarSplitscreen": "TAGame.GameEvent_SoccarSplitscreen_TA",
    "Archetypes.SpecialPickups.SpecialPickup_BallFreeze": "TAGame.SpecialPickup_BallFreeze_TA",
    "Archetypes.SpecialPickups.SpecialPickup_BallGrapplingHook": "TAGame.SpecialPickup_GrapplingHook_TA",
    "Archetypes.SpecialPickups.SpecialPickup_BallLasso": "TAGame.SpecialPickup_BallLasso_TA",
    "Archetypes.SpecialPickups.SpecialPickup_BallSpring": "TAGame.SpecialPickup_BallCarSpring_TA",
    "Archetypes.SpecialPickups.SpecialPickup_BallVelcro": "TAGame.SpecialPickup_BallVelcro_TA",
    "Archetypes.SpecialPickups.SpecialPickup_Batarang": "TAGame.SpecialPickup_Batarang_TA",
    "Archetypes.SpecialPickups.SpecialPickup_BoostOverride": "TAGame.SpecialPickup_BoostOverride_TA",
    "Archetypes.SpecialPickups.SpecialPickup_CarSpring": "TAGame.SpecialPickup_BallCarSpring_TA",
    "Archetypes.SpecialPickups.SpecialPickup_GravityWell": "TAGame.SpecialPickup_BallGravity_TA",
    "Archetypes.SpecialPickups.SpecialPickup_StrongHit": "TAGame.SpecialPickup_HitForce_TA",
    "Archetypes.SpecialPickups.SpecialPickup_Swapper": "TAGame.SpecialPickup_Swapper_TA",
    "Archetypes.SpecialPickups.SpecialPickup_Tornado": "TAGame.SpecialPickup_Tornado_TA",
    "Archetypes.SpecialPickups.SpecialPickup_HauntedBallBeam": "TAGame.SpecialPickup_HauntedBallBeam_TA",
    "Archetypes.SpecialPickups.SpecialPickup_Rugby": "TAGame.SpecialPickup_Rugby_TA",
    "Archetypes.Teams.Team0": "TAGame.Team_Soccar_TA",
    "Archetypes.Teams.Team1": "TAGame.Team_Soccar_TA",
    "GameInfo_Basketball.GameInfo.GameInfo_Basketball:GameReplicationInfoArchetype": "TAGame.GRI_TA",
    "GameInfo_Breakout.GameInfo.GameInfo_Breakout:GameReplicationInfoArchetype": "TAGame.GRI_TA",
    "Gameinfo_Hockey.GameInfo.Gameinfo_Hockey:GameReplicationInfoArchetype": "TAGame.GRI_TA",
    "GameInfo_Items.GameInfo.GameInfo_Items:GameReplicationInfoArchetype": "TAGame.GRI_TA",
    "GameInfo_Season.GameInfo.GameInfo_Season:GameReplicationInfoArchetype": "TAGame.GRI_TA",
    "GameInfo_Soccar.GameInfo.GameInfo_Soccar:GameReplicationInfoArchetype": "TAGame.GRI_TA",
    "GameInfo_Tutorial.GameInfo.GameInfo_Tutorial:GameReplicationInfoArchetype": "TAGame.GRI_TA",
    "GameInfo_Tutorial.GameEvent.GameEvent_Tutorial_Aerial": "TAGame.GameEvent_Tutorial_TA",
    "TAGame.Default__CameraSettingsActor_TA": "TAGame.CameraSettingsActor_TA",
    "TAGame.Default__PRI_TA": "TAGame.PRI_TA",
    "TAGame.Default__Car_TA": "TAGame.Car_TA",
    "TheWorld:PersistentLevel.BreakOutActor_Platform_TA": "TAGame.BreakOutActor_Platform_TA",
    "TheWorld:PersistentLevel.CrowdActor_TA": "TAGame.CrowdActor_TA",
    "TheWorld:PersistentLevel.CrowdManager_TA": "TAGame.CrowdManager_TA",
    "TheWorld:PersistentLevel.InMapScoreboard_TA": "TAGame.InMapScoreboard_TA",
    "TheWorld:PersistentLevel.VehiclePickup_Boost_TA": "TAGame.VehiclePickup_Boost_TA",
    "Haunted_TrainStation_P.TheWorld:PersistentLevel.HauntedBallTrapTrigger_TA_1": "TAGame.HauntedBallTrapTrigger_TA",
    "Haunted_TrainStation_P.TheWorld:PersistentLevel.HauntedBallTrapTrigger_TA_0": "TAGame.HauntedBallTrapTrigger_TA",
    "Archetypes.Tutorial.Cannon": "TAGame.Cannon_TA"
}



from tempfile import NamedTemporaryFile

import pytest

from carball.analysis.analysis_manager import AnalysisManager
from carball.tests.utils import run_analysis_test_on_replay, get_raw_replays


class TestExport:

    def test_json_export(self, replay_cache):
        def test(analysis: AnalysisManager):
            with NamedTemporaryFile(mode='w') as f:
                analysis.write_json_out_to_file(f)

        run_analysis_test_on_replay(test, get_raw_replays()["DEFAULT_3_ON_3_AROUND_58_HITS"], cache=replay_cache)

    def test_proto_export(self, replay_cache):
        def test(analysis: AnalysisManager):
            with NamedTemporaryFile(mode='wb') as f:
                analysis.write_proto_out_to_file(f)

        run_analysis_test_on_replay(test, get_raw_replays()["DEFAULT_3_ON_3_AROUND_58_HITS"], cache=replay_cache)

    def test_unicode_names(self, replay_cache):
        def test(analysis: AnalysisManager):
            with NamedTemporaryFile(mode='wb') as f:
                analysis.write_pandas_out_to_file(f)

        run_analysis_test_on_replay(test, get_raw_replays()["UNICODE_ERROR"], cache=replay_cache)

    def test_json_export_invalid_type(self, replay_cache):
        def test(analysis: AnalysisManager):
            with NamedTemporaryFile(mode='wb') as f:
                with pytest.raises(IOError):
                    analysis.write_json_out_to_file(f)

        run_analysis_test_on_replay(test, get_raw_replays()["DEFAULT_3_ON_3_AROUND_58_HITS"], cache=replay_cache)

    def test_proto_export_invalid_type(self, replay_cache):
        def test(analysis: AnalysisManager):
            with NamedTemporaryFile(mode='w') as f:
                with pytest.raises(IOError):
                    analysis.write_proto_out_to_file(f)

        run_analysis_test_on_replay(test, get_raw_replays()["DEFAULT_3_ON_3_AROUND_58_HITS"], cache=replay_cache)

    def test_unicode_names_invalid_type(self, replay_cache):
        def test(analysis: AnalysisManager):
            with NamedTemporaryFile(mode='w') as f:
                with pytest.raises(IOError):
                    analysis.write_pandas_out_to_file(f)

        run_analysis_test_on_replay(test, get_raw_replays()["UNICODE_ERROR"], cache=replay_cache)



import math
from typing import Dict

RBSTATE = "TAGame.RBActor_TA:ReplicatedRBState"

RB_STATE_DICT_PAIRS = {
    'pos_x': (RBSTATE, 'location', 'x'),
    'pos_y': (RBSTATE, 'location', 'y'),
    'pos_z': (RBSTATE, 'location', 'z'),
    'quat_w': (RBSTATE, 'rotation', 'w'),
    'quat_x': (RBSTATE, 'rotation', 'x'),
    'quat_y': (RBSTATE, 'rotation', 'y'),
    'quat_z': (RBSTATE, 'rotation', 'z'),
    'vel_x': (RBSTATE, 'linear_velocity', 'x'),
    'vel_y': (RBSTATE, 'linear_velocity', 'y'),
    'vel_z': (RBSTATE, 'linear_velocity', 'z'),
    'ang_vel_x': (RBSTATE, 'angular_velocity', 'x'),
    'ang_vel_y': (RBSTATE, 'angular_velocity', 'y'),
    'ang_vel_z': (RBSTATE, 'angular_velocity', 'z'),
}

BALL_DATA_DICT_PAIRS = {
    **RB_STATE_DICT_PAIRS,
    'hit_team_no': ("TAGame.Ball_TA:HitTeamNum",),
}


class BallActor:
    @staticmethod
    def get_data_dict(actor_data: Dict) -> Dict:
        # is_sleeping = actor_data.get(RBSTATE, {}).get(rbstate, {}).get('sleeping', True)
        data_dict = get_data_dict_from_pairs(actor_data, pairs=BALL_DATA_DICT_PAIRS)
        data_dict = standardise_data_dict(data_dict)
        return data_dict


# these are parsed such that the max values are as follow:
# pos_x: 3088, pos_y: 5424, pos_z: 1826
# mag(vel) = 23000
# mag(ang_vel) = 5500

CAR_DATA_DICT_PAIRS = {
    **RB_STATE_DICT_PAIRS,
    'throttle': ("TAGame.Vehicle_TA:ReplicatedThrottle",),
    'steer': ("TAGame.Vehicle_TA:ReplicatedSteer",),
    'handbrake': ("TAGame.Vehicle_TA:bReplicatedHandbrake",),
}


class CarActor:
    @staticmethod
    def get_data_dict(actor_data: Dict) -> Dict:
        # is_sleeping = actor_data.get(RBSTATE, {}).get(rbstate, {}).get('sleeping', True)
        data_dict = get_data_dict_from_pairs(actor_data, pairs=CAR_DATA_DICT_PAIRS)
        data_dict = standardise_data_dict(data_dict)
        return data_dict


def get_data_dict_from_pairs(actor_data: dict, pairs: dict) -> dict:
    data_dict = {}

    for _key, _value_keys in pairs.items():
        _value = actor_data
        for _value_key in _value_keys:
            _value = _value.get(_value_key, None)
            if _value is None:
                break
        data_dict[_key] = _value
    return data_dict


def standardise_data_dict(data_dict: dict) -> dict:
    data_dict = rescale_to_uu(data_dict)
    if data_dict['quat_w'] is not None:
        data_dict = convert_quat_to_rot(data_dict)
    data_dict.pop('quat_w')
    data_dict.pop('quat_x')
    data_dict.pop('quat_y')
    data_dict.pop('quat_z')
    return data_dict


def rescale_to_uu(data_dict: dict) -> dict:
    # handle psyonix's rounding to 2dp (and storing 1.00 as 100)
    correction_dict = {'vel_x': 0.1, 'vel_y': 0.1, 'vel_z': 0.1,
                       'ang_vel_x': 0.1, 'ang_vel_y': 0.1, 'ang_vel_z': 0.1}
    # /100 pos, /10 vel and ang_vel
    for _item, _divisor in correction_dict.items():
        try:
            data_dict[_item] /= _divisor
        except TypeError:
            continue
    return data_dict


def convert_quat_to_rot(data_dict: dict) -> dict:
    # https://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles#Quaternion_to_Euler_Angles_Conversion

    w = data_dict['quat_w']
    x = data_dict['quat_x']
    y = data_dict['quat_y']
    z = data_dict['quat_z']

    sinr = 2.0 * (w * x + y * z)
    cosr = 1.0 - 2.0 * (x * x + y * y)
    roll = math.atan2(sinr, cosr)

    sinp = 2.0 * (w * y - z * x)
    if abs(sinp) >= 1:
        pitch = math.copysign(math.pi / 2, sinp)
    else:
        pitch = math.asin(sinp)

    siny = 2.0 * (w * z + x * y)
    cosy = 1.0 - 2.0 * (y * y + z * z)
    yaw = math.atan2(siny, cosy)
    data_dict['rot_x'] = pitch
    data_dict['rot_y'] = yaw
    data_dict['rot_z'] = roll

    return data_dict



import logging
from typing import Dict

import pandas as pd
import numpy as np

from carball.analysis.stats.stats import BaseStat
from carball.generated.api.player_pb2 import Player
from carball.generated.api import game_pb2
from carball.json_parser.game import Game
from carball.analysis.stats.dropshot import is_dropshot

log = logging.getLogger(__name__)


class DropshotBallPhaseTimes(BaseStat):
    def calculate_stat(self, proto_stat, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                       data_frame: pd.DataFrame):
        if not is_dropshot(game):
            return

        phase_air_times = [[], [], []]

        all_events = list(map(lambda x: (x['frame_number'], x['state']), game.dropshot['ball_events']))
        all_events.extend(map(lambda x: (x['frame_number'], 0), game.dropshot['damage_events']))
        all_events.sort(key=lambda x: x[0])

        current_frame = data_frame['ball'].index[0]
        current_phase = 0

        for ball_event in all_events:
            if ball_event[1] == current_phase:
                continue

            phase_time = data_frame['game', 'time'].loc[ball_event[0]] - data_frame['game', 'time'].loc[current_frame]
            phase_air_times[current_phase].append(phase_time)

            current_frame = ball_event[0]
            current_phase = ball_event[1]

        # last phase
        phase_time = data_frame['game', 'time'].iloc[-1] - data_frame['game', 'time'].loc[current_frame]
        phase_air_times[current_phase].append(phase_time)

        for phase, times in enumerate(phase_air_times):
            phase_stat_proto = proto_game.game_stats.ball_stats.extra_mode.dropshot_phase_stats.add()
            phase_stat_proto.phase = phase
            if len(times) > 0:
                phase_stat_proto.average = np.mean(times)
                phase_stat_proto.max = max(times)
                phase_stat_proto.total = sum(times)
            else:
                phase_stat_proto.average = 0
                phase_stat_proto.max = 0
                phase_stat_proto.total = 0



def initialize_project():
    print("Importing necessary methods...")
    from utils.create_proto import create_proto_files
    from utils.import_fixer import convert_to_relative_imports

    print("Creating proto files...")
    create_proto_files()
    convert_to_relative_imports()


if __name__ == "__main__":
    print("Initialising project...")
    initialize_project()



import json
import os

import setuptools
from setuptools import setup
from setuptools.command.develop import develop
from setuptools.command.install import install

with open(os.path.join('carball', 'analysis', 'PROTOBUF_VERSION'), 'r') as f:
    PROTOBUF_VERSION = json.loads(f.read())


with open(os.path.join('CARBALL_VERSION'), 'r') as f:
    subversion = json.loads(f.read())

version_string = '0.' + str(PROTOBUF_VERSION) + '.' + str(subversion)

if os.path.isfile('README.md'):
    with open("README.md", "r") as readme_file:
        long_description = readme_file.read()
else:
    long_description = ''


class PostDevelopCommand(develop):
    """Post-installation for development mode."""

    def run(self):
        from init import initialize_project
        initialize_project()
        # this needs to be last
        develop.run(self)


class PostInstallCommand(install):
    """Post-installation for installation mode."""

    def run(self):
        from init import initialize_project
        initialize_project()
        # this needs to be last
        install.run(self)


setup(
    name='carball',
    version=version_string,
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=['pandas==0.24.2', 'protobuf==3.6.1', 'xlrd==1.1.0', 'numpy==1.17.0', 'boxcars-py==0.1.3'],
    url='https://github.com/SaltieRL/carball',
    keywords=['rocket-league'],
    license='Apache 2.0',
    author='Matthew Mage, Harry Xie, David Turner',
    author_email='sciguymjm@gmail.com',
    description='Rocket League replay parsing and analysis.',
    long_description=long_description,
    exclude_package_data={'': ['.gitignore', '.git/*', '.git/**/*', 'replays/*']},
    long_description_content_type='text/markdown',
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand,
    },
    entry_points={
        'console_scripts': ['carball=carball.command_line:main']
    }
)



import unittest

from carball.analysis.analysis_manager import AnalysisManager

from carball.tests.utils import run_analysis_test_on_replay, get_specific_replays


class DBTest(unittest.TestCase):
    def test_error_replay(self):
        local = self

        def test(analysis: AnalysisManager):
            local.assertIsNone(analysis.get_protobuf_data())
            local.assertEqual(False, analysis.get_protobuf_data().game_metadata.is_invalid_analysis)

        try:
            run_analysis_test_on_replay(test, replay_list=get_specific_replays()["BROKEN_REPLAYS"])
        except FileNotFoundError as e:
            self.assertIsInstance(e, FileNotFoundError)



from .base import *


class GameInfoHandler(BaseActorHandler):

    @classmethod
    def can_handle(cls, actor: dict) -> bool:
        return actor['TypeName'].endswith(':GameReplicationInfoArchetype')

    def update(self, actor: dict, frame_number: int, time: float, delta: float) -> None:
        if self.parser.game_info_actor is None:
            self.parser.game_info_actor = actor



from carball.tests.utils import get_replay_path
from carball.decompile_replays import analyze_replay_file

"""
This file is doing benchmarking for the main replay varieties. The tests are run via the benchmarking GitHub action.

kwargs is a dict of variable name to value for the benchmarked method.
rounds are collections of iterations, and the data is collected off of these.
iterations are single method runs.
"""


def test_short_sample(benchmark):
    replay_path = get_replay_path("SHORT_SAMPLE.replay")

    benchmark.pedantic(analyze_replay_file,
                       kwargs={"replay_path":replay_path}, rounds=10, iterations=3)


def test_short_dropshot(benchmark):
    replay_path = get_replay_path("DROPSHOT_PHASE2_BALL.replay")

    benchmark.pedantic(analyze_replay_file,
                       kwargs={"replay_path":replay_path}, rounds=10, iterations=3)


def test_full_rumble(benchmark):
    replay_path = get_replay_path("RUMBLE_FULL.replay")

    benchmark.pedantic(analyze_replay_file,
                       kwargs={"replay_path":replay_path}, rounds=10, iterations=3)


def test_oce_rlcs(benchmark):
    replay_path = get_replay_path("OCE_RLCS_7_CARS.replay")

    benchmark.pedantic(analyze_replay_file,
                       kwargs={"replay_path":replay_path}, rounds=10, iterations=3)


def test_intensive_oce_rlcs(benchmark):
    replay_path = get_replay_path("OCE_RLCS_7_CARS.replay")

    benchmark.pedantic(analyze_replay_file,
                       kwargs={"replay_path":replay_path, "calculate_intensive_events":True}, rounds=5, iterations=1)



import math
import logging
from typing import Dict

import pandas as pd

from carball.analysis.stats.stats import BaseStat
from carball.generated.api.player_pb2 import Player
from carball.generated.api import game_pb2
from carball.json_parser.game import Game
from carball.analysis.stats.dropshot import is_dropshot
from carball.analysis.constants.dropshot import *

log = logging.getLogger(__name__)


class DropshotGoals(BaseStat):
    def calculate_stat(self, proto_stat, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                       data_frame: pd.DataFrame):
        if not is_dropshot(game):
            return

        tile_positions = get_tile_positions(game.map)

        if tile_positions is None:
            log.warning(f'Unsupported dropshot map: {game.map}')
            return

        for goal in proto_game.game_metadata.goals:
            frame = goal.frame_number

            # get the ball position at goal
            ball_pos = data_frame['ball'].loc[frame].loc[['pos_x', 'pos_y', 'pos_z']]

            # get the closest tile
            team = player_map[goal.player_id.id].is_orange
            closest_distance = TILE_DIAMETER
            closest_id = -1

            opponent_tile_ids = get_team_tiles(game.map, team ^ 1)
            opponent_tiles = [tile_positions[i] for i in opponent_tile_ids]

            for tile_id, tile in enumerate(opponent_tiles):

                d = math.sqrt(
                    math.pow(ball_pos['pos_x'] - tile[0], 2) +
                    math.pow(ball_pos['pos_y'] - tile[1], 2) +
                    math.pow(ball_pos['pos_z'] - tile[2], 2)
                )

                if d < closest_distance:
                    closest_distance = d
                    closest_id = tile_id if team == 1 else tile_id + 70

            if closest_id != -1:
                goal.extra_mode_info.dropshot_tile.id = closest_id

            # find the previous damage frame
            damage_frame = max(filter(lambda x: x <= frame, game.dropshot['tile_frames'].keys()))

            # get the team tile states
            tile_states = {k: v for (k, v) in game.dropshot['tile_frames'][damage_frame].items() if k in opponent_tile_ids}

            # get number of damaged tiles
            damaged_tiles = sum(1 for _ in filter(lambda x: x == 1, tile_states.values()))
            destroyed_tiles = sum(1 for _ in filter(lambda x: x == 2, tile_states.values()))

            goal.extra_mode_info.phase_1_tiles = damaged_tiles
            goal.extra_mode_info.phase_2_tiles = destroyed_tiles



import logging
from typing import Callable

from carball.generated.api import party_pb2
from ....json_parser.game import Game
from ....generated.api.metadata import game_metadata_pb2
from .ApiDemo import ApiDemo
from .ApiGoal import ApiGoal

log = logging.getLogger(__name__)


class ApiGameScore:
    @staticmethod
    def create_from_game(game):
        game_score = game_metadata_pb2.GameScore()
        for team in game.teams:
            if team.is_orange:
                game_score.team_1_score = team.score
            else:
                game_score.team_0_score = team.score
        return game_score


class ApiGame:
    @staticmethod
    def create_from_game(proto_game: game_metadata_pb2.GameMetadata,
                         game: Game, id_creator: Callable) -> game_metadata_pb2.GameMetadata:
        proto_game.id = game.id
        proto_game.name = str(game.name)
        proto_game.map = game.map
        if game.replay_version is not None:
            proto_game.version = game.replay_version
        proto_game.time = int(game.datetime.timestamp())
        proto_game.frames = game.frames.index.max()
        proto_game.score.CopyFrom(ApiGameScore.create_from_game(game))
        ApiGoal.create_goals_from_game(game, proto_game.goals, id_creator)
        ApiDemo.create_demos_from_game(game, proto_game.demos, id_creator)
        if game.primary_player is not None and game.primary_player['id'] is not None:
            proto_game.primary_player.id = game.primary_player['id']
        try:
            proto_game.playlist = game.game_info.playlist
        except:
            proto_game.playlist = 0
            try:
                proto_game.unknown_playlist = int(game.game_info.playlist)
            except:
                log.error('unknown and invalid playlist')

        proto_game.game_server_id = game.game_info.server_id
        proto_game.server_name = game.game_info.server_name
        if game.game_info.match_guid is None or game.game_info.match_guid == '':
            proto_game.match_guid = game.id
        else:
            proto_game.match_guid = game.game_info.match_guid
        proto_game.team_size = game.team_size

        return id_creator, proto_game

    @staticmethod
    def create_parties(parties, game: Game, id_creator: Callable):
        for k, v in game.parties.items():
            proto_party: party_pb2.Party = parties.add()
            proto_party.leader_id.id = k
            for p in v:
                player = proto_party.members.add()
                player.id = p



import numpy as np
import pandas as pd


def sum_deltas_by_player_name(data_frame: pd.DataFrame, players_data_frame: pd.Series) -> pd.DataFrame:
    """
    Gets the delta from the pandas data frame for certain players at certain times
    :param data_frame: The game data frame, goal frames are removed
    :param players_data_frame: Just the players
    :return: The time based on delta.
    """
    combined_data = pd.concat([players_data_frame, data_frame['game', 'delta'].rename('delta')], axis=1)
    return combined_data.groupby(players_data_frame.name).sum().rename(columns={'delta': players_data_frame.name})


def sum_deltas_start_end_frame(data_frame: pd.DataFrame, start_frame, end_frame) -> pd.DataFrame:
    """
    Gets the delta from the pandas data frame for a given start and end frame
    :param data_frame: The game data frame, goal frames are removed
    :param start_frame: When to start counting the delta
    :param end_frame: When to stop counting the delta
    :return: The time based on delta.
    """
    return data_frame['game', 'delta'][start_frame: end_frame].sum()


def sum_deltas_by_truthy_data(data_frame: pd.DataFrame, truthy_frames: pd.Series) -> np.float64:
    """
    Gets the delta from the pandas data frame for certain players at certain times.
    :param data_frame: The game data frame, goal frames are removed
    :param truthy_frames: Frames that have a truth value applied to them.
    :return: The time based on delta.
    """
    truthy_frames = truthy_frames.rename('truthy')
    combined_data = pd.concat([truthy_frames, data_frame['game', 'delta'].rename('delta')], axis=1)
    return combined_data.loc[combined_data['truthy'] == True].sum().loc['delta']


import glob
import gzip
import os
import shutil
import subprocess
import traceback
from io import BytesIO

from google.protobuf.json_format import MessageToJson

from carball.analysis.utils.pandas_manager import PandasManager
from carball.controls.controls import ControlsCreator
from carball.decompile_replays import analyze_replay_file
from carball.json_parser.sanity_check.sanity_check import SanityChecker


def create_dir(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)


def __test_replays(BASE_DIR):
    import logging

    ROOT_DIR = os.path.dirname(BASE_DIR)
    OUTPUT_DIR = os.path.join(ROOT_DIR, 'output')

    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)
    MOVE_WORKING = True
    DEBUGGING = True
    success = 0
    failure = 0
    create_dir(OUTPUT_DIR)
    sanity_check = SanityChecker()
    sanity_check = None

    for filepath in glob.iglob(ROOT_DIR + '/**carball/replays/*.replay', recursive=True):
        logger.info('decompiling %s', filepath)
        if "output" in filepath:
            continue
        base_file = os.path.basename(filepath)
        json_path = os.path.join(OUTPUT_DIR, 'replays/decompiled/{}'.format(base_file.replace(".replay", ".json")))
        proto_path = os.path.join(OUTPUT_DIR, 'replays/protos/{}'.format(base_file.replace(".replay", ".pts")))
        pandas_path = os.path.join(OUTPUT_DIR, 'replays/pandas/{}'.format(base_file.replace(".replay", ".gzip")))

        create_dir(os.path.dirname(json_path))
        create_dir(os.path.dirname(proto_path))
        create_dir(os.path.dirname(pandas_path))

        if DEBUGGING:
            try:
                analysis_manager = analyze_replay_file(filepath, json_path,
                                                       controls=ControlsCreator(), analysis_per_goal=False,
                                                       sanity_check=sanity_check)
                id = analysis_manager.protobuf_game.game_metadata.id
                with open(os.path.join(OUTPUT_DIR, id + 'game.json'), 'w') as f:
                    f.write(MessageToJson(analysis_manager.protobuf_game))
                    bytes = PandasManager.write_pandas_to_buffer_for_tooling(analysis_manager.get_data_frame(),
                                                           analysis_manager.get_protobuf_data().players)
                    with open(pandas_path, 'wb') as fo:
                        fo.write(bytes)

                data_frame = PandasManager.safe_read_pandas_to_memory(BytesIO(analysis_manager.df_bytes))
                logger.info('length of decoded pandas %i', len(data_frame))
            except subprocess.CalledProcessError as e:
                traceback.print_exc()
        else:
            try:
                analysis_manager = analyze_replay_file(filepath, json_path)
                with open(proto_path, 'wb') as fo:
                    analysis_manager.write_proto_out_to_file(fo)
                with gzip.open(pandas_path, 'wb') as fo:
                    analysis_manager.write_pandas_out_to_file(fo)
                if MOVE_WORKING:
                    shutil.move(filepath, os.path.join('replays', 'working', filepath))
                success += 1
            except Exception as e:
                traceback.print_exc()
                failure += 1
    if not DEBUGGING:
        if float(success + failure) == 0:
            print("NO REPLAYS WERE RUN.")
            print("Need files in: " + BASE_DIR)
        ratio = success / float(success + failure)
        print('success ratio:', ratio)

if __name__ == "__main__":
    __test_replays(os.path.dirname(os.getcwd()))



from .ball import *


class TeamHandler(BaseActorHandler):

    @classmethod
    def can_handle(cls, actor: dict) -> bool:
        return actor['ClassName'] == 'TAGame.Team_Soccar_TA'

    def update(self, actor: dict, frame_number: int, time: float, delta: float) -> None:
        self.parser.team_dicts[actor['Id']] = actor
        self.parser.team_dicts[actor['Id']]['colour'] = 'blue' if actor["TypeName"] == "Archetypes.Teams.Team0" else \
            'orange'



import numpy as np
import pandas as pd
from carball.generated.api import game_pb2
from carball.analysis.constants.field_constants import FieldConstants


class PickupAnalysis:
    field_constants = FieldConstants()
    BIG_BOOST_POSITIONS = field_constants.get_big_pads()
    SMALL_BOOST_POSITIONS = field_constants.get_small_pads()
    BIG_BOOST_RADIUS = 208
    SMALL_BOOST_RADIUS = 149  # 144 doesn't work for some pickups that are very close to the edge.
    BIG_BOOST_HEIGHT = 168
    SMALL_BOOST_HEIGHT = 165
    # Choosing how many frames to be open to setting a pickup. Back is for when the player is ahead of the server (usually smaller)
    LAG_BACK = 6
    LAG_FORWARD = 14

    @classmethod
    def add_pickups(cls, proto_game: game_pb2.Game, data_frame: pd.DataFrame):

        for player in proto_game.players:
            player_vals_df = data_frame[player.name][['pos_x', 'pos_y', 'pos_z', 'boost']].copy()
            player_vals_df['boost'] /= 2.55
            player_vals_df['boost'] = player_vals_df['boost'].round(5)
            player_vals_df = player_vals_df.dropna(axis=0, how='all')
            player_vals_df = player_vals_df.fillna(0)
            player_vals_df['boost_collect'] = cls.get_boost_collect(player_vals_df)
            data_frame[player.name, 'boost_collect'] = player_vals_df['boost_collect']
        return

    @classmethod
    def get_boost_collect(cls, player_vals_df):
        # Get a series with indexes as a subset of the indexes of df, values being pad label picked up.
        # Iterate through every pad, label each frame in the path with which boost pad it was in range of.
        df = player_vals_df.copy()
        path = df.drop(['pos_z', 'boost'], axis=1)
        big_labels = np.zeros(len(path))
        small_labels = np.zeros(len(path))
        # Calculate the distances from each pad. Add label of the pad if distance <= radius
        for pad in cls.BIG_BOOST_POSITIONS:
            distances = np.sqrt(np.square(path.values - pad[:2]).sum(axis=1, dtype=np.float32))
            big_labels += (pad[2] * (distances <= cls.BIG_BOOST_RADIUS))

        for pad in cls.SMALL_BOOST_POSITIONS:
            distances = np.sqrt(np.square(path.values - pad[:2]).sum(axis=1, dtype=np.float32))
            small_labels += (pad[2] * (distances <= cls.SMALL_BOOST_RADIUS))
        # Add labels and exclude labels with z too high. Didn't calculate this earlier because its a flat height)
        df['pad_in_range'] = 0
        df['pad_in_range'] += small_labels
        df.loc[df['pos_z'] >= cls.SMALL_BOOST_HEIGHT, 'pad_in_range'] = 0
        df['pad_in_range'] += big_labels
        df.loc[df['pos_z'] >= cls.BIG_BOOST_HEIGHT, 'pad_in_range'] = 0
        # Get the gains in boost per frame
        df['gains'] = df['boost'].diff().clip(0)
        # Get whether we entered or exited the range of a pad per frame
        df['status_change'] = (df['pad_in_range'].diff(1))
        df = df.fillna(0)
        # Get the index of the frame we most recently entered a pad range, per frame.
        df['recent_entry_index'] = df.index
        df.loc[df['status_change'] <= 0, 'recent_entry_index'] = 0
        df['recent_entry_index'] = df['recent_entry_index'].replace(0, np.nan).fillna(
            method='bfill', limit=cls.LAG_BACK).fillna(
            method='ffill', limit=cls.LAG_FORWARD)
        gains_frames = df.loc[
            ((df['gains'] > 5) & (df['boost'] != 33.33333)) | ((df['gains'] > 0) & (df['boost'] > 95.0))].copy()

        gains_indexes = gains_frames['recent_entry_index'].dropna()

        pickups = df.loc[gains_indexes]['status_change'].copy()
        pickups = pickups.loc[~pickups.index.duplicated(keep='first')]
        return pickups



import unittest

from carball.analysis.analysis_manager import AnalysisManager

from carball.tests.utils import run_analysis_test_on_replay, get_specific_replays, get_specific_answers, \
    assertNearlyEqual, get_raw_replays


class Test_Boost():

    def test_1_small_pad_collected(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            frames = analysis.get_data_frame()
            player = proto_game.players[0]
            boost = player.stats.boost
            assert(boost.num_small_boosts == 1)

        run_analysis_test_on_replay(test, get_specific_replays()["1_SMALL_PAD"], cache=replay_cache)

    def test_1_large_pad_collected(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            player = proto_game.players[0]
            boost = player.stats.boost
            assert(boost.num_large_boosts == 1)

        run_analysis_test_on_replay(test, get_specific_replays()["1_LARGE_PAD"], cache=replay_cache)

    def test_1_large_pad_1_small_pad_collected(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            player = proto_game.players[0]
            boost = player.stats.boost
            assert(boost.num_large_boosts == 1)
            assert(boost.num_small_boosts == 1)

        run_analysis_test_on_replay(test, get_raw_replays()["12_AND_100_BOOST_PADS_0_USED"], cache=replay_cache)

    def test_0_boost_collected(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            player = proto_game.players[0]
            boost = player.stats.boost
            assert(boost.num_small_boosts == 0)
            assert(boost.num_large_boosts == 0)

        run_analysis_test_on_replay(test, get_specific_replays()["0_BOOST_COLLECTED"], cache=replay_cache)

    def test_lots_of_boost_collected(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            player = proto_game.players[0]
            boost = player.stats.boost
            assert [boost.num_small_boosts, boost.num_large_boosts] == [25, 6]

        run_analysis_test_on_replay(test, get_raw_replays()["6_BIG_25_SMALL"], cache=replay_cache)

    def test_boost_steals(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            player = proto_game.players[0]
            boost = player.stats.boost
            assert boost.num_stolen_boosts == 2

        run_analysis_test_on_replay(test, get_raw_replays()["6_BIG_25_SMALL"], cache=replay_cache)

    def test_boost_steals_post_goal(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            player = proto_game.players[0]
            boost = player.stats.boost
            assert [boost.num_small_boosts, boost.num_large_boosts,
                    boost.num_stolen_boosts, boost.boost_usage] == [0, 0, 0, 0]

            player = proto_game.players[1]
            boost = player.stats.boost
            assert [boost.num_large_boosts, boost.num_stolen_boosts] == [3, 3]
            assert boost.boost_usage > 0

        run_analysis_test_on_replay(test, get_raw_replays()["3_STEAL_ORANGE_0_STEAL_BLUE"], cache=replay_cache)

    def test_boost_used(self, replay_cache):
        case = unittest.TestCase('__init__')

        def test(analysis: AnalysisManager, boost_value):
            proto_game = analysis.get_protobuf_data()
            player = proto_game.players[0]
            boost = player.stats.boost
            print("Predicted usage: {}, actual: {}".format(boost.boost_usage, boost_value))
            case.assertAlmostEqual(boost.boost_usage, boost_value, delta=1)
            # self.assertGreater(boost.average_boost_level, 0)

        run_analysis_test_on_replay(test, get_specific_replays()["BOOST_USED"] + get_specific_replays()["0_BOOST_USED"],
                                    answers=get_specific_answers()["BOOST_USED"] +
                                            get_specific_answers()["0_BOOST_USED"], cache=replay_cache)

    def test_boost_feathered(self, replay_cache):
        case = unittest.TestCase('__init__')

        def test(analysis: AnalysisManager, boost_value):
            proto_game = analysis.get_protobuf_data()
            player = proto_game.players[0]
            boost = player.stats.boost
            print("Predicted usage: {}, actual: {}".format(boost.boost_usage, boost_value))
            assertNearlyEqual(case, boost.boost_usage, boost_value, percent=3)
            # self.assertGreater(boost.average_boost_level, 0)

        run_analysis_test_on_replay(test, get_specific_replays()["BOOST_FEATHERED"],
                                    answers=get_specific_answers()["BOOST_FEATHERED"], cache=replay_cache)

    def test_boost_wasted_collection(self, replay_cache):
        case = unittest.TestCase('__init__')

        def test(analysis: AnalysisManager, boost_value):
            proto_game = analysis.get_protobuf_data()
            for index, player in enumerate(proto_game.players):
                wasted_answer = boost_value[index]
                total_wasted = (wasted_answer[0] - (255 - wasted_answer[1])) / 256.0 * 100.0
                boost = player.stats.boost
                case.assertAlmostEqual(boost.wasted_collection, total_wasted, delta=2)

        run_analysis_test_on_replay(test, get_specific_replays()["BOOST_WASTED_COLLECTION"],
                                    answers=get_specific_answers()["BOOST_WASTED_COLLECTION"], cache=replay_cache)

    def test_0_used(self, replay_cache):
        def test(analysis: AnalysisManager, boost_value):
            proto_game = analysis.get_protobuf_data()
            player = proto_game.players[0]
            boost = player.stats.boost
            assert(boost.boost_usage == boost_value)
            print(analysis)

        run_analysis_test_on_replay(test, get_specific_replays()["0_BOOST_USED"],
                                    answers=get_specific_answers()["0_BOOST_USED"], cache=replay_cache)



from typing import List, Dict
import pandas as pd
from ....analysis.stats.stats import BaseStat

from ....generated.api import game_pb2
from ....generated.api.player_pb2 import Player
from ....generated.api.stats.events_pb2 import Hit
from ....generated.api.stats.player_stats_pb2 import PlayerStats
from ....json_parser.game import Game


class Averages(BaseStat):
    def calculate_player_stat(self, player_stat_map: Dict[str, PlayerStats], game: Game, proto_game: game_pb2.Game,
                              player_map: Dict[str, Player], data_frame: pd.DataFrame):

        for key, player in player_map.items():
            self.get_averages_for_player(player, proto_game, data_frame)

    @classmethod
    def get_averages_for_player(cls, player: Player, proto_game: game_pb2.Game, data_frame: pd.DataFrame):
        player_hits: List[Hit] = [saltie_hit for saltie_hit in proto_game.game_stats.hits if
                                  saltie_hit.player_id == player.id]

        hit_distances = [saltie_hit.distance for saltie_hit in player_hits
                         if saltie_hit.distance is not None and not saltie_hit.dribble]
        if len(hit_distances) > 0:
            average_hit_distance = sum(hit_distances) / len(hit_distances)
            player.stats.averages.average_hit_distance = average_hit_distance



import logging
from typing import Dict

import pandas as pd

from carball.analysis.stats.stats import BaseStat
from carball.generated.api.player_pb2 import Player
from carball.generated.api import game_pb2
from carball.generated.api.stats.player_stats_pb2 import PlayerStats
from carball.generated.api.stats.team_stats_pb2 import TeamStats
from carball.json_parser.game import Game
from carball.analysis.stats.dropshot import is_dropshot

log = logging.getLogger(__name__)


class DropshotStats(BaseStat):

    def calculate_player_stat(self, player_stat_map: Dict[str, PlayerStats], game: Game, proto_game: game_pb2.Game,
                              player_map: Dict[str, Player], data_frame: pd.DataFrame):
        if not is_dropshot(game):
            return

        player_stats = {}

        for key, val in player_map.items():
            player_stats[key] = {
                'total': 0,
                'max': 0
            }

        for event in game.dropshot['damage_events']:
            ball_phase = data_frame['ball', 'dropshot_phase'].loc[event['frame_number'] - 1]
            max_dmg = 1
            if ball_phase == 1:
                max_dmg = 7
            elif ball_phase == 2:
                max_dmg = 19
            player_id = str(event['player'].online_id)
            player_stats[player_id]['total'] += len(event['tiles'])
            player_stats[player_id]['max'] += max_dmg

        self.apply_damage_stats(player_stat_map, player_stats)

    def calculate_team_stat(self, team_stat_list: Dict[int, TeamStats], game: Game, proto_game: game_pb2.Game,
                            player_map: Dict[str, Player], data_frame: pd.DataFrame):
        if not is_dropshot(game):
            return

        team_stats = {
            0: {'total': 0, 'max': 0},
            1: {'total': 0, 'max': 0}
        }

        for event in game.dropshot['damage_events']:
            ball_phase = data_frame['ball', 'dropshot_phase'].loc[event['frame_number'] - 1]
            max_dmg = 1
            if ball_phase == 1:
                max_dmg = 7
            elif ball_phase == 2:
                max_dmg = 19

            team = event['player'].is_orange
            team_stats[team]['total'] += len(event['tiles'])
            team_stats[team]['max'] += max_dmg

        self.apply_damage_stats(team_stat_list, team_stats)

    def calculate_stat(self, proto_stat, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                       data_frame: pd.DataFrame):
        if not is_dropshot(game):
            return

        tile_stats = {}
        damaged = 0
        destroyed = 0

        for event in game.dropshot['damage_events']:
            for tile_damage in event['tiles']:
                tile_id = tile_damage[0]
                if tile_id not in tile_stats:
                    tile_stats[tile_id] = 1
                else:
                    tile_stats[tile_id] += 1

                if tile_damage[1] == 1:
                    damaged += 1
                elif tile_damage[1] == 2:
                    destroyed += 1

        for tile_id, total_damage in tile_stats.items():
            damage_stat_proto = proto_game.game_stats.dropshot_stats.tile_stats.damage_stats.add()
            damage_stat_proto.id = tile_id
            damage_stat_proto.total_damage = total_damage

        proto_game.game_stats.dropshot_stats.tile_stats.damaged_tiles = damaged
        proto_game.game_stats.dropshot_stats.tile_stats.destroyed_tiles = destroyed

    def apply_damage_stats(self, stat_list, game_stats):
        for key, stats in stat_list.items():
            stats.dropshot_stats.total_damage = game_stats[key]['total']

            if game_stats[key]['max'] > 0:
                stats.dropshot_stats.damage_efficiency = game_stats[key]['total'] / game_stats[key]['max']
            else:
                stats.dropshot_stats.damage_efficiency = 0
